from __future__ import annotations

from bcsfe import core


class MapNames:
    def __init__(
        self,
        save_file: core.SaveFile,
        code: str,
        base_index: int,
        output: bool = True,
        no_r_prefix: bool = False,
    ):
        self.save_file = save_file
        self.out = output
        self.code = code
        self.base_index = base_index
        self.no_r_prefix = no_r_prefix
        self.gdg = core.core_data.get_game_data_getter(self.save_file)
        self.map_names: dict[int, str | None] = {}
        self.stage_names: dict[int, list[str]] = {}
        self.get_map_names()

    def get_map_names_in_game(
        self, base_index: int, total_stages: int
    ) -> dict[int, str | None] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        map_name_data = gdg.download("resLocal", "Map_Name.csv")
        if map_name_data is None:
            return None

        csv = core.CSV(
            map_name_data, core.Delimeter.from_country_code_res(self.save_file.cc)
        )
        names: dict[int, str | None] = {}
        for row in csv:
            id = row[0].to_int()
            name = row[1].to_str().strip()

            for i in range(total_stages):
                index = i + base_index
                if id == index:
                    if name:
                        names[i] = name
                    else:
                        names[i] = None
                    break

        return names

    def get_map_names(self) -> dict[int, str | None] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        r_prefix = "" if self.no_r_prefix else "R"
        stage_names = gdg.download(
            "resLocal",
            f"StageName_{r_prefix}{self.code}_{core.core_data.get_lang(self.save_file)}.csv",
        )
        if stage_names is None:
            return None
        csv = core.CSV(
            stage_names,
            core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        for i, row in enumerate(csv):
            stage_names_row = row.to_str_list()
            if not stage_names_row:
                continue
            self.stage_names[i] = stage_names_row

        names = self.get_map_names_in_game(self.base_index, len(self.stage_names))
        if names is None:
            return None
        self.map_names = names
        return self.map_names

    @staticmethod
    def get_code_from_id(id: int) -> str | None:
        base_id = id // 1000

        ids = {
            0: "RN",
            1: "RS",
            2: "RC",
            4: "EX",
            6: "RT",
            7: "RV",
            11: "RR",
            12: "RM",
            13: "RNA",
            14: "RB",
            16: "RD",
            20: "Z",
            21: "Z",
            22: "Z",
            24: "RA",
            25: "RH",
            27: "RCA",
            30: "DM",
            31: "RQ",
            32: "L",
            34: "RND",
        }

        return ids.get(base_id)

    @staticmethod
    def from_id(id: int, save_file: core.SaveFile) -> MapNames | None:
        code = MapNames.get_code_from_id(id)
        if code is None:
            return None
        return MapNames(save_file, code, id, no_r_prefix=True)



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class UncannyChapters:
    def __init__(self, chapters: core.Chapters, unknown: list[int]):
        self.chapters = chapters
        self.unknown = unknown

    @staticmethod
    def init() -> UncannyChapters:
        return UncannyChapters(core.Chapters.init(), [])

    @staticmethod
    def read(data: core.Data) -> UncannyChapters:
        ch = core.Chapters.read(data, read_every_time=False)
        unknown = data.read_int_list(length=len(ch.chapters))
        return UncannyChapters(ch, unknown)

    def write(self, data: core.Data):
        self.chapters.write(data, write_every_time=False)
        data.write_int_list(self.unknown, write_length=False)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": self.chapters.serialize(),
            "unknown": self.unknown,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> UncannyChapters:
        return UncannyChapters(
            core.Chapters.deserialize(data.get("chapters", {})),
            data.get("unknown", []),
        )

    def __repr__(self):
        return f"Uncanny({self.chapters}, {self.unknown})"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def edit_uncanny(save_file: core.SaveFile):
        uncanny = save_file.uncanny
        uncanny.chapters.edit_chapters(save_file, "NA", 13000)

    @staticmethod
    def edit_catamin_stages(save_file: core.SaveFile):
        choice = dialog_creator.ChoiceInput.from_reduced(
            ["change_clear_amount_catamin", "clear_unclear_stage_catamin"],
            dialog="catamin_stage_clear_q",
        ).single_choice()
        if choice is None:
            return None

        if choice == 1:
            names = core.MapNames(save_file, "B", base_index=14000)
            map_ids = core.EventChapters.select_map_names(names.map_names)
            if map_ids is None:
                return None
            if len(map_ids) >= 2:
                choice2 = dialog_creator.ChoiceInput.from_reduced(
                    ["individual", "all_at_once"], dialog="catamin_clear_amounts_q"
                ).single_choice()
                if choice2 is None:
                    return None
            else:
                choice2 = 1

            if choice2 == 2:
                clear_amount = dialog_creator.IntInput().get_input(
                    "enter_clear_amount_catamin", {}
                )[0]
                if clear_amount is None:
                    return None
                for map_id in map_ids:
                    save_file.event_stages.chapter_completion_count[14_000 + map_id] = (
                        clear_amount
                    )
            elif choice == 1:
                for map_id in map_ids:
                    name = names.map_names.get(map_id) or core.localize("unknown_map")
                    clear_amount = dialog_creator.IntInput().get_input(
                        "enter_clear_amount_catamin_map", {"name": name, "id": map_id}
                    )[0]
                    if clear_amount is None:
                        return None
                    save_file.event_stages.chapter_completion_count[14_000 + map_id] = (
                        clear_amount
                    )

            color.ColoredText.localize("catamin_stage_success")

        elif choice == 2:
            completed_chapters = save_file.catamin_stages.chapters.edit_chapters(
                save_file, "B", 14000
            )
            if completed_chapters is None:
                return None

            # TODO: maybe in the future ask if the user wants to modify the chapter clear amounts



from __future__ import annotations
from collections.abc import Callable
from typing import Type, TypeVar

from bcsfe import core


class FilterDate:
    def __init__(self, start_mmdd: int, start_hhmm: int, end_mmdd: int, end_hhmm: int):
        self.start_mmdd = start_mmdd
        self.start_hhmm = start_hhmm
        self.end_mmdd = end_mmdd
        self.end_hhmm = end_hhmm

    @staticmethod
    def from_csv_row(row: core.Row) -> FilterDate:
        return FilterDate(
            row.next_int(), row.next_int(), row.next_int(), row.next_int()
        )


class FilterItem:
    def __init__(
        self,
        filter_date: FilterDate | None,
        filter_day_flags: list[bool],  # 31 item array
        filter_week: int,
        filter_times_start_end_hhmm: list[tuple[int, int]],
    ):
        self.filter_date = filter_date
        self.filter_day_flags = filter_day_flags
        self.filter_week = filter_week
        self.filter_times_start_end_hhmm = filter_times_start_end_hhmm

    @staticmethod
    def from_csv_row(row: core.Row) -> FilterItem:
        filter_date_enabled = row.next_bool()

        filter_date = None
        if filter_date_enabled:
            filter_date = FilterDate.from_csv_row(row)

        filter_day_count = row.next_int()

        filter_day_flags: list[bool] = [False] * 31

        for _ in range(filter_day_count):
            day_ind = row.next_int() - 1
            if day_ind >= 0 and day_ind < len(filter_day_flags):
                filter_day_flags[day_ind] = True

        filter_week = row.next_int()
        filter_time_count = row.next_int()

        filter_times_start_end_hhmm: list[tuple[int, int]] = []

        for _ in range(filter_time_count):
            start_hhmm = row.next_int()
            end_hhmm = row.next_int()

            filter_times_start_end_hhmm.append((start_hhmm, end_hhmm))

        return FilterItem(
            filter_date, filter_day_flags, filter_week, filter_times_start_end_hhmm
        )


def split_yyyymmdd(yyyymmdd: int) -> tuple[int, int, int]:
    year = yyyymmdd // 10_000
    month = (yyyymmdd % 10_000) // 100
    day = yyyymmdd % 100

    return year, month, day


def split_hhmm(hhmm: int) -> tuple[int, int]:
    hour = hhmm // 100
    minute = hhmm % 100

    return hour, minute


class FilterData:
    def __init__(
        self,
        start_yyyymmdd: int,
        start_hhmm: int,
        end_yyyymmdd: int,
        end_hhmm: int,
        min_game_version: int,
        max_game_version: int,
        platform_flag: int,
        filter_items: list[FilterItem],
    ):
        self.start_yyyymmdd = start_yyyymmdd
        self.start_hhmm = start_hhmm
        self.end_yyyymmdd = end_yyyymmdd
        self.end_hhmm = end_hhmm
        self.min_game_version = min_game_version
        self.max_game_version = max_game_version
        self.platform_flag = platform_flag
        self.filter_items = filter_items

    @staticmethod
    def from_csv_row(row: core.Row) -> FilterData:
        start_yyyymmdd = row.next_int()
        start_hhmm = row.next_int()
        end_yyyymmdd = row.next_int()
        end_hhmm = row.next_int()
        min_game_version = row.next_int()
        max_game_version = row.next_int()
        platform_flag = row.next_int()
        total_items = row.next_int()

        filter_items: list[FilterItem] = []

        for _ in range(total_items):
            filter_items.append(FilterItem.from_csv_row(row))

        return FilterData(
            start_yyyymmdd,
            start_hhmm,
            end_yyyymmdd,
            end_hhmm,
            min_game_version,
            max_game_version,
            platform_flag,
            filter_items,
        )


class Localization:
    def __init__(self, lang: str, title: str, message: str):
        self.lang = lang
        self.title = title
        self.message = message

    @staticmethod
    def from_csv_row(row: core.Row) -> Localization:
        return Localization(row.next_str(), row.next_str(), row.next_str())


class RarityGatya:
    def __init__(self, prob: int, guaranteed: int):
        self.prob = prob
        self.guaranteed = guaranteed

    @staticmethod
    def from_csv_row(row: core.Row) -> RarityGatya:
        return RarityGatya(row.next_int(), row.next_int())


class ServerGatyaDataSet:
    def __init__(
        self,
        number: int,
        catfood: int,
        stage_progress: int,
        flags: int,
        rarity_info: list[RarityGatya],
        message: str,
        collab_message: tuple[str, str] | None,
    ):
        self.number = number
        self.catfood = catfood
        self.stage_progress = stage_progress
        self.flags = flags
        self.rarity_info = rarity_info
        self.message = message
        self.other_event_message = collab_message

    @staticmethod
    def from_csv_row(row: core.Row, flag: int) -> ServerGatyaDataSet:
        number = row.next_int()
        catfood = row.next_int()
        stage_progress = row.next_int()
        flags = row.next_int()
        rarity_info: list[RarityGatya] = []

        for _ in range(5):
            rarity_info.append(RarityGatya.from_csv_row(row))

        message = row.next_str()

        collab_message = None
        if flag == 4:
            collab_message = (row.next_str(), row.next_str())

        return ServerGatyaDataSet(
            number,
            catfood,
            stage_progress,
            flags,
            rarity_info,
            message,
            collab_message,
        )

    def is_visible_silhouette(self) -> bool:
        return (self.flags & 1) != 0

    def is_required_user_rank_1600(self) -> bool:
        return (self.flags & 2) != 0

    def has_stepup_gatya(self) -> bool:
        return (self.flags & 4) != 0


class ServerGatyaDataItem:
    def __init__(self, filter: FilterData, flags: int, sets: list[ServerGatyaDataSet]):
        self.filter = filter
        self.flags = flags
        self.sets = sets

    @staticmethod
    def from_csv_row(row: core.Row) -> ServerGatyaDataItem:
        filter = FilterData.from_csv_row(row)
        flag = row.next_int()
        count = row.next_int()

        sets: list[ServerGatyaDataSet] = []

        for _ in range(count):
            sets.append(ServerGatyaDataSet.from_csv_row(row, flag))

        return ServerGatyaDataItem(filter, flag, sets)

    def get_normal_flag(self) -> bool:
        return self.flags == 0

    def get_rare_flag(self) -> bool:
        return 1 <= self.flags <= 3

    def get_collab_flag(self) -> bool:
        return self.flags == 4

    def get_first_rare_flag(self) -> bool:
        return self.flags == 2

    def get_first_rare_10_flag(self) -> bool:
        return self.flags == 3


class ServerItemDataItem:
    def __init__(
        self,
        filter: FilterData,
        event_number: int,  # server item id
        item_number: int,
        item_unit: int,  # base quanity, not cat unit (e.g 2 XP+1000s)
        title: str,
        message: str,
        stage_progress: int,
        stage_progress_flag: int,
        flags: int,
        locales: list[Localization] | None,
    ):
        self.filter = filter
        self.event_number = event_number
        self.item_number = item_number
        self.item_unit = item_unit
        self.title = title
        self.message = message
        self.stage_progress = stage_progress
        self.stage_progress_flag = stage_progress_flag
        self.flags = flags
        self.locales = locales

    def is_every_day(self) -> bool:
        return (self.flags & 1) != 0

    def is_required_user_rank_1600(self) -> bool:
        return (self.flags & 2) != 0

    @staticmethod
    def from_csv_row(row: core.Row) -> ServerItemDataItem:
        filter = FilterData.from_csv_row(row)

        event_number = row.next_int()
        item_number = row.next_int()
        item_unit = row.next_int()
        title = row.next_str()
        message = row.next_str()
        stage_progress = row.next_int()
        stage_progress_flag = row.next_bool()
        flags = row.next_int()

        locales: list[Localization] | None = None

        if not row.done():
            locales = []
            total_locales = row.next_int()

            for _ in range(total_locales):
                locales.append(Localization.from_csv_row(row))

        return ServerItemDataItem(
            filter,
            event_number,
            item_number,
            item_unit,
            title,
            message,
            stage_progress,
            stage_progress_flag,
            flags,
            locales,
        )


Item = TypeVar("Item")
T = TypeVar("T")


def read_event_data(
    csv: core.CSV,
    read_func: Callable[[core.Row], Item],
    init_func: Callable[[list[Item]], T],
) -> T | None:
    start = csv.read_line()
    if start is None:
        return None

    if start.next_str() != "[start]":
        return None

    if not start.done():
        return None

    items: list[Item] = []

    while True:
        row = csv.read_line()
        if row is None:
            return None

        if len(row) == 0:
            return None

        if row[0].to_str() == "[end]":
            break

        item = read_func(row)

        items.append(item)

    return init_func(items)


class ServerItemData:
    def __init__(self, items: list[ServerItemDataItem]):
        self.items = items

    @staticmethod
    def from_csv(csv: core.CSV) -> ServerItemData | None:
        return read_event_data(csv, ServerItemDataItem.from_csv_row, ServerItemData)

    @staticmethod
    def from_data(data: core.Data) -> ServerItemData | None:
        csv = core.CSV(data, delimiter="\t", remove_comments=False, remove_empty=False)

        return ServerItemData.from_csv(csv)


class ServerGatyaData:
    def __init__(self, items: list[ServerGatyaDataItem]):
        self.items = items

    @staticmethod
    def from_csv(csv: core.CSV) -> ServerGatyaData | None:
        return read_event_data(csv, ServerGatyaDataItem.from_csv_row, ServerGatyaData)

    @staticmethod
    def from_data(data: core.Data) -> ServerGatyaData | None:
        csv = core.CSV(data, delimiter="\t", remove_comments=False, remove_empty=False)

        return ServerGatyaData.from_csv(csv)



from __future__ import annotations
from bcsfe import core


class MapResetData:
    def __init__(
        self,
        yearly_end_timestamp: float,
        monthly_end_timestamp: float,
        weekly_end_timestamp: float,
        daily_end_timestamp: float,
    ):
        self.yearly_end_timestamp = yearly_end_timestamp
        self.monthly_end_timestamp = monthly_end_timestamp
        self.weekly_end_timestamp = weekly_end_timestamp
        self.daily_end_timestamp = daily_end_timestamp

    @staticmethod
    def init() -> MapResetData:
        return MapResetData(
            0.0,
            0.0,
            0.0,
            0.0,
        )

    @staticmethod
    def read(stream: core.Data) -> MapResetData:
        yearly_end_timestamp = stream.read_double()
        monthly_end_timestamp = stream.read_double()
        weekly_end_timestamp = stream.read_double()
        daily_end_timestamp = stream.read_double()
        return MapResetData(
            yearly_end_timestamp,
            monthly_end_timestamp,
            weekly_end_timestamp,
            daily_end_timestamp,
        )

    def write(self, stream: core.Data):
        stream.write_double(self.yearly_end_timestamp)
        stream.write_double(self.monthly_end_timestamp)
        stream.write_double(self.weekly_end_timestamp)
        stream.write_double(self.daily_end_timestamp)

    def serialize(self) -> dict[str, float]:
        return {
            "yearly_end_timestamp": self.yearly_end_timestamp,
            "monthly_end_timestamp": self.monthly_end_timestamp,
            "weekly_end_timestamp": self.weekly_end_timestamp,
            "daily_end_timestamp": self.daily_end_timestamp,
        }

    @staticmethod
    def deserialize(data: dict[str, float]) -> MapResetData:
        return MapResetData(
            data.get("yearly_end_timestamp", 0.0),
            data.get("monthly_end_timestamp", 0.0),
            data.get("weekly_end_timestamp", 0.0),
            data.get("daily_end_timestamp", 0.0),
        )

    def __str__(self) -> str:
        return f"MapResetData(yearly_end_timestamp={self.yearly_end_timestamp!r}, monthly_end_timestamp={self.monthly_end_timestamp!r}, weekly_end_timestamp={self.weekly_end_timestamp!r}, daily_end_timestamp={self.daily_end_timestamp!r})"

    def __repr__(self) -> str:
        return str(self)


class MapResets:
    def __init__(self, data: dict[int, list[MapResetData]]):
        self.data = data

    @staticmethod
    def init() -> MapResets:
        return MapResets({})

    @staticmethod
    def read(stream: core.Data) -> MapResets:
        data: dict[int, list[MapResetData]] = {}
        for _ in range(stream.read_int()):
            key = stream.read_int()
            value: list[MapResetData] = []
            for _ in range(stream.read_int()):
                value.append(MapResetData.read(stream))
            data[key] = value
        return MapResets(data)

    def write(self, stream: core.Data):
        stream.write_int(len(self.data))
        for key, value in self.data.items():
            stream.write_int(key)
            stream.write_int(len(value))
            for item in value:
                item.write(stream)

    def serialize(self) -> dict[int, list[dict[str, float]]]:
        return {
            key: [item.serialize() for item in value]
            for key, value in self.data.items()
        }

    @staticmethod
    def deserialize(data: dict[int, list[dict[str, float]]]) -> MapResets:
        return MapResets(
            {
                key: [MapResetData.deserialize(item) for item in value]
                for key, value in data.items()
            }
        )

    def __str__(self) -> str:
        return f"MapResets(data={self.data!r})"

    def __repr__(self) -> str:
        return str(self)



from __future__ import annotations
from bcsfe import core
import tempfile


class PackageNameNotSet(Exception):
    pass


class RootHandler:
    def __init__(self):
        self.package_name = None

    def is_android(self) -> bool:
        return core.Path.get_root().add("system").exists()

    def set_package_name(self, package_name: str):
        self.package_name = package_name

    def is_rooted(self) -> bool:
        try:
            core.Path.get_root().add("data").add("data").get_dirs()
        except PermissionError:
            return False
        return True

    def get_battlecats_packages(self) -> list[str]:
        packages = core.Path.get_root().add("data").add("data").get_dirs()
        packages = [
            package.basename()
            for package in packages
            if package.add("files").add("SAVE_DATA").exists()
        ]
        return packages

    def get_package_name(self) -> str:
        if self.package_name is None:
            raise PackageNameNotSet("Package name is not set")
        return self.package_name

    def get_battlecats_path(self) -> core.Path:
        return core.Path.get_root().add("data").add("data").add(self.get_package_name())

    def get_battlecats_save_path(self) -> core.Path:
        return self.get_battlecats_path().add("files").add("SAVE_DATA")

    def save_battlecats_save(self, local_path: core.Path) -> core.CommandResult:
        self.get_battlecats_save_path().copy(local_path)
        return core.CommandResult.create_success()

    def load_battlecats_save(self, local_path: core.Path) -> core.CommandResult:
        local_path.copy(self.get_battlecats_save_path())
        return core.CommandResult.create_success()

    def close_game(self) -> core.CommandResult:
        cmd = core.Command(
            f"sudo pkill -f {self.get_package_name()}",
        )
        return cmd.run()

    def run_game(self) -> core.CommandResult:
        cmd = core.Command(
            f"sudo monkey -p {self.get_package_name()} -c android.intent.category.LAUNCHER 1",
        )
        return cmd.run()

    def rerun_game(self) -> core.CommandResult:
        result = self.close_game()
        if not result.success:
            return result
        result = self.run_game()
        if not result.success:
            return result

        return core.CommandResult.create_success()

    def save_locally(
        self, local_path: core.Path | None = None
    ) -> tuple[core.Path | None, core.CommandResult]:
        if local_path is None:
            local_path = core.Path.get_documents_folder().add("saves").add("SAVE_DATA")
        local_path.parent().generate_dirs()
        result = self.save_battlecats_save(local_path)
        if not result.success:
            return None, result

        return local_path, result

    def load_locally(self, local_path: core.Path) -> core.CommandResult:
        success = self.load_battlecats_save(local_path)
        if not success:
            return core.CommandResult.create_failure()

        success = self.rerun_game()
        if not success:
            return core.CommandResult.create_failure()

        return core.CommandResult.create_success()

    def load_save(
        self, save: core.SaveFile, rerun_game: bool = True
    ) -> core.CommandResult:
        with tempfile.TemporaryDirectory() as temp_dir:
            local_path = core.Path(temp_dir).add("SAVE_DATA")
            save.to_data().to_file(local_path)
            result = self.load_battlecats_save(local_path)
            if not result.success:
                return result
        if rerun_game:
            result = self.rerun_game()

        return result



from __future__ import annotations
from typing import Any

from requests.exceptions import ConnectionError
from requests import Response
from json.decoder import JSONDecodeError
from bcsfe.cli import color, dialog_creator

from bcsfe.core import (
    country_code,
    crypto,
    game,
    game_version,
    io,
    locale_handler,
    log,
    server,
    theme_handler,
    max_value_helper,
)
from bcsfe.core.country_code import CountryCode, CountryCodeType
from bcsfe.core.crypto import Hash, HashAlgorithm, Hmac, NyankoSignature, Random
from bcsfe.core.game.battle.battle_items import BattleItems
from bcsfe.core.game.battle.cleared_slots import ClearedSlots
from bcsfe.core.game.battle.slots import LineUps
from bcsfe.core.game.battle.enemy import (
    Enemy,
    EnemyNames,
    EnemyDescriptions,
    EnemyDictionary,
)
from bcsfe.core.game.catbase.beacon_base import BeaconEventListScene
from bcsfe.core.game.catbase.cat import (
    Cat,
    Cats,
    UnitBuy,
    TalentData,
    NyankoPictureBook,
    StorageItem,
)
from bcsfe.core.game.catbase.gambling import GamblingEvent
from bcsfe.core.game.catbase.gatya import (
    Gatya,
    GatyaInfos,
    GatyaDataSet,
    GatyaDataOptionSet,
    GatyaDataOption,
)
from bcsfe.core.game.catbase.gatya_item import (
    GatyaItemBuy,
    GatyaItemNames,
    GatyaItemCategory,
    GatyaItemBuyItem,
)
from bcsfe.core.game.catbase.item_pack import (
    ItemPack,
    Purchases,
    PurchaseSet,
    PurchasedPack,
)
from bcsfe.core.game.catbase.login_bonuses import LoginBonus
from bcsfe.core.game.catbase.matatabi import Matatabi
from bcsfe.core.game.catbase.drop_chara import CharaDrop
from bcsfe.core.game.catbase.medals import Medals, MedalNames
from bcsfe.core.game.catbase.mission import (
    Missions,
    MissionNames,
    MissionConditions,
)
from bcsfe.core.game.catbase.my_sale import MySale
from bcsfe.core.game.catbase.nyanko_club import NyankoClub
from bcsfe.core.game.catbase.officer_pass import OfficerPass
from bcsfe.core.game.catbase.powerup import PowerUpHelper
from bcsfe.core.game.catbase.scheme_items import SchemeItems
from bcsfe.core.game.catbase.special_skill import (
    SpecialSkills,
    SpecialSkill,
    AbilityData,
    AbilityDataItem,
)

from bcsfe.core.game.catbase.stamp import StampData
from bcsfe.core.game.catbase.talent_orbs import (
    TalentOrb,
    TalentOrbs,
    OrbInfo,
    OrbInfoList,
    RawOrbInfo,
    SaveOrb,
    SaveOrbs,
)
from bcsfe.core.game.catbase.unlock_popups import (
    UnlockPopups,
    UnlockPopupData,
    UnlockPopupLine,
)
from bcsfe.core.game.catbase.upgrade import Upgrade
from bcsfe.core.game.catbase.user_rank_rewards import (
    UserRankRewards,
    RankGifts,
    RankGiftDescriptions,
)
from bcsfe.core.game.catbase.playtime import PlayTime
from bcsfe.core.game.gamoto.base_materials import BaseMaterials
from bcsfe.core.game.gamoto.cat_shrine import CatShrine, CatShrineLevels
from bcsfe.core.game.gamoto.gamatoto import (
    Gamatoto,
    GamatotoLevels,
    GamatotoMembersName,
)
from bcsfe.core.game.gamoto.ototo import Ototo
from bcsfe.core.game.localizable import Localizable
from bcsfe.core.game.map.aku import AkuChapters
from bcsfe.core.game.map.challenge import ChallengeChapters
from bcsfe.core.game.map.chapters import Chapters
from bcsfe.core.game.map.dojo import Dojo
from bcsfe.core.game.map.enigma import Enigma
from bcsfe.core.game.map.event import EventChapters
from bcsfe.core.game.map.ex_stage import ExChapters
from bcsfe.core.game.map.gauntlets import GauntletChapters
from bcsfe.core.game.map.item_reward_stage import ItemRewardChapters
from bcsfe.core.game.map.legend_quest import LegendQuestChapters
from bcsfe.core.game.map.map_reset import MapResets
from bcsfe.core.game.map.outbreaks import Outbreaks
from bcsfe.core.game.map.story import StoryChapters, TreasureText, StageNames
from bcsfe.core.game.map.timed_score import TimedScoreChapters
from bcsfe.core.game.map.tower import TowerChapters
from bcsfe.core.game.map.uncanny import UncannyChapters
from bcsfe.core.game.map.zero_legends import ZeroLegendsChapters
from bcsfe.core.game.map.map_names import MapNames
from bcsfe.core.game.map.map_option import MapOption
from bcsfe.core.game_version import GameVersion
from bcsfe.core.io.adb_handler import AdbHandler, AdbNotInstalled
from bcsfe.core.io.waydroid import WayDroidHandler
from bcsfe.core.io.bc_csv import CSV, Delimeter, Row
from bcsfe.core.io.command import Command, CommandResult
from bcsfe.core.io.config import Config, ConfigKey
from bcsfe.core.io.data import Data
from bcsfe.core.io.json_file import JsonFile
from bcsfe.core.io.path import Path
from bcsfe.core.io.save import SaveError, SaveFile, CantDetectSaveCCError
from bcsfe.core.io.thread_helper import thread_run_many, Thread
from bcsfe.core.io.yaml import YamlFile
from bcsfe.core.io.git_handler import GitHandler, Repo
from bcsfe.core.io.root_handler import RootHandler
from bcsfe.core.locale_handler import (
    LocalManager,
    ExternalLocaleManager,
    ExternalLocale,
)
from bcsfe.core.log import Logger
from bcsfe.core.server.event_data import (
    ServerItemData,
    ServerItemDataItem,
    ServerGatyaData,
    ServerGatyaDataSet,
    ServerGatyaDataItem,
)
from bcsfe.core.server.client_info import ClientInfo
from bcsfe.core.server.game_data_getter import GameDataGetter
from bcsfe.core.server.headers import AccountHeaders
from bcsfe.core.server.managed_item import (
    BackupMetaData,
    ManagedItem,
    ManagedItemType,
)
from bcsfe.core.server.request import RequestHandler, MultiPartFile, MultipartForm
from bcsfe.core.server.server_handler import ServerHandler
from bcsfe.core.server.updater import Updater
from bcsfe.core.theme_handler import (
    ThemeHandler,
    ExternalTheme,
    ExternalThemeManager,
)
from bcsfe.core.max_value_helper import MaxValueHelper, MaxValueType


class CoreData:
    def init_data(self):
        self.config = Config(config_path, print_config_err)
        self.logger = Logger(log_path)
        self.local_manager = LocalManager()
        self.theme_manager = ThemeHandler()
        self.max_value_manager = MaxValueHelper()
        self.game_data_getter: GameDataGetter | None = None
        self.gatya_item_names: GatyaItemNames | None = None
        self.gatya_item_buy: GatyaItemBuy | None = None
        self.chara_drop: CharaDrop | None = None
        self.gamatoto_levels: GamatotoLevels | None = None
        self.gamatoto_members_name: GamatotoMembersName | None = None
        self.localizable: Localizable | None = None
        self.abilty_data: AbilityData | None = None
        self.enemy_names: EnemyNames | None = None
        self.rank_gift_descriptions: RankGiftDescriptions | None = None
        self.rank_gifts: RankGifts | None = None
        self.treasure_text: TreasureText | None = None
        self.cat_shrine_levels: CatShrineLevels | None = None
        self.medal_names: MedalNames | None = None
        self.mission_names: MissionNames | None = None
        self.mission_conditions: MissionConditions | None = None

    def get_game_data_getter(
        self,
        save: SaveFile | None = None,
        cc: CountryCode | None = None,
        gv: GameVersion | None = None,
    ) -> GameDataGetter:
        if self.game_data_getter is None:
            if cc is None and save is not None:
                cc = save.cc
            if cc is None:
                raise ValueError("cc must be provided if save is not provided")
            if gv is None and save is not None:
                gv = save.game_version
            if gv is None:
                raise ValueError("gv must be provided if save is not provided")
            self.game_data_getter = GameDataGetter(cc, gv)
        return self.game_data_getter

    def get_gatya_item_names(self, save: SaveFile) -> GatyaItemNames:
        if self.gatya_item_names is None:
            self.gatya_item_names = GatyaItemNames(save)
        return self.gatya_item_names

    def get_gatya_item_buy(self, save: SaveFile) -> GatyaItemBuy:
        if self.gatya_item_buy is None:
            self.gatya_item_buy = GatyaItemBuy(save)
        return self.gatya_item_buy

    def get_chara_drop(self, save: SaveFile) -> CharaDrop:
        if self.chara_drop is None:
            self.chara_drop = CharaDrop(save)
        return self.chara_drop

    def get_gamatoto_levels(self, save: SaveFile) -> GamatotoLevels:
        if self.gamatoto_levels is None:
            self.gamatoto_levels = GamatotoLevels(save)
        return self.gamatoto_levels

    def get_gamatoto_members_name(self, save: SaveFile) -> GamatotoMembersName:
        if self.gamatoto_members_name is None:
            self.gamatoto_members_name = GamatotoMembersName(save)
        return self.gamatoto_members_name

    def get_localizable(self, save: SaveFile) -> Localizable:
        if self.localizable is None:
            self.localizable = Localizable(save)
        return self.localizable

    def get_ability_data(self, save: SaveFile) -> AbilityData:
        if self.abilty_data is None:
            self.abilty_data = AbilityData(save)
        return self.abilty_data

    def get_enemy_names(self, save: SaveFile) -> EnemyNames:
        if self.enemy_names is None:
            self.enemy_names = EnemyNames(save)
        return self.enemy_names

    def get_rank_gift_descriptions(self, save: SaveFile) -> RankGiftDescriptions:
        if self.rank_gift_descriptions is None:
            self.rank_gift_descriptions = RankGiftDescriptions(save)
        return self.rank_gift_descriptions

    def get_rank_gifts(self, save: SaveFile) -> RankGifts:
        if self.rank_gifts is None:
            self.rank_gifts = RankGifts(save)
        return self.rank_gifts

    def get_treasure_text(self, save: SaveFile) -> TreasureText:
        if self.treasure_text is None:
            self.treasure_text = TreasureText(save)
        return self.treasure_text

    def get_cat_shrine_levels(self, save: SaveFile) -> CatShrineLevels:
        if self.cat_shrine_levels is None:
            self.cat_shrine_levels = CatShrineLevels(save)
        return self.cat_shrine_levels

    def get_medal_names(self, save: SaveFile) -> MedalNames:
        if self.medal_names is None:
            self.medal_names = MedalNames(save)
        return self.medal_names

    def get_mission_names(self, save: SaveFile) -> MissionNames:
        if self.mission_names is None:
            self.mission_names = MissionNames(save)
        return self.mission_names

    def get_mission_conditions(self, save: SaveFile) -> MissionConditions:
        if self.mission_conditions is None:
            self.mission_conditions = MissionConditions(save)
        return self.mission_conditions

    def get_lang(self, save: SaveFile) -> str:
        return self.get_localizable(save).get_lang() or "en"


config_path = None
print_config_err = True
log_path = None


def set_config_path(path: Path):
    global config_path
    config_path = path


def set_log_path(path: Path):
    global log_path
    log_path = path


def update_external_content(_: Any = None):
    """Updates external content."""

    color.ColoredText.localize("updating_external_content")
    print()
    ExternalThemeManager.update_all_external_themes()
    ExternalLocaleManager.update_all_external_locales()
    core_data.init_data()

    clear_game_data = dialog_creator.YesNoInput().get_input_once("clear_game_data_q")
    if clear_game_data is None:
        return

    if clear_game_data:
        GameDataGetter.delete_old_versions(0)
        color.ColoredText.localize("cleared_game_data")


def print_no_internet():
    color.ColoredText.localize("no_internet")


core_data = CoreData()


def localize(key: str, escape: bool = True, **kwargs: Any) -> str:
    return core_data.local_manager.get_key(key, escape=escape, **kwargs)


__all__ = [
    "server",
    "io",
    "locale_handler",
    "country_code",
    "log",
    "game_version",
    "crypto",
    "game",
    "theme_handler",
    "max_value_helper",
    "AdbHandler",
    "AdbNotInstalled",
    "CountryCode",
    "Path",
    "Data",
    "CSV",
    "ServerHandler",
    "GameVersion",
    "SaveFile",
    "JsonFile",
    "ManagedItem",
    "ManagedItemType",
    "BackupMetaData",
    "Cat",
    "Upgrade",
    "PowerUpHelper",
    "TalentOrb",
    "TalentOrbs",
    "OrbInfo",
    "OrbInfoList",
    "RawOrbInfo",
    "SaveOrb",
    "SaveOrbs",
    "ConfigKey",
    "SpecialSkill",
    "WayDroidHandler",
    "EnemyDescriptions",
    "EnemyDictionary",
    "GatyaItemCategory",
    "ServerItemData",
    "GatyaItemBuyItem",
    "ServerItemDataItem",
    "ServerGatyaData",
    "ServerGatyaDataSet",
    "ServerGatyaDataItem",
    "GatyaDataOptionSet",
    "GatyaDataOption",
    "MaxValueType",
    "GamblingEvent",
    "UnitBuy",
    "NyankoPictureBook",
    "StorageItem",
    "OfficerPass",
    "LocalManager",
    "MapOption",
    "CantDetectSaveCCError",
    "UnlockPopupData",
    "UnlockPopupLine",
]



from __future__ import annotations
import json
from typing import Any
from bcsfe import core


class JsonFile:
    def __init__(self, data: core.Data):
        self.json = json.loads(data.data)

    @staticmethod
    def from_path(path: core.Path) -> JsonFile:
        return JsonFile(path.read())

    @staticmethod
    def from_object(js: Any) -> JsonFile:
        return JsonFile(core.Data(json.dumps(js)))

    @staticmethod
    def from_data(data: core.Data) -> JsonFile:
        return JsonFile(data)

    def to_data(self, indent: int | None = 4) -> core.Data:
        return core.Data(json.dumps(self.json, indent=indent))

    def to_file(self, path: core.Path) -> None:
        path.write(self.to_data())

    def to_object(self) -> Any:
        return self.json

    def get(self, key: str) -> Any:
        return self.json[key]

    def set(self, key: str, value: Any) -> None:
        self.json[key] = value

    def __str__(self) -> str:
        return str(self.json)

    def __getitem__(self, key: str) -> Any:
        return self.json[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.json[key] = value



from bcsfe.core.game.catbase import (
    gatya_item,
    stamp,
    cat,
    upgrade,
    special_skill,
    my_sale,
    gatya,
    user_rank_rewards,
    item_pack,
    login_bonuses,
    scheme_items,
    unlock_popups,
    beacon_base,
    mission,
    nyanko_club,
    officer_pass,
    medals,
    talent_orbs,
    matatabi,
    powerup,
    drop_chara,
    playtime,
    gambling,
)

__all__ = [
    "stamp",
    "cat",
    "upgrade",
    "special_skill",
    "my_sale",
    "gatya",
    "user_rank_rewards",
    "item_pack",
    "login_bonuses",
    "scheme_items",
    "unlock_popups",
    "beacon_base",
    "mission",
    "nyanko_club",
    "officer_pass",
    "medals",
    "talent_orbs",
    "gatya_item",
    "matatabi",
    "powerup",
    "drop_chara",
    "playtime",
    "gambling",
]



from __future__ import annotations
from bcsfe import core


class Enemy:
    def __init__(self, id: int):
        self.id = id

    def unlock_enemy_guide(self, save_file: core.SaveFile):
        save_file.enemy_guide[self.id] = 1

    def reset_enemy_guide(self, save_file: core.SaveFile):
        save_file.enemy_guide[self.id] = 0

    def get_name(self, save_file: core.SaveFile) -> str | None:
        return core.core_data.get_enemy_names(save_file).get_name(self.id)


class EnemyDictionaryItem:
    def __init__(self, enemy_id: int, scale: int, first_seen: int | None):
        self.enemy_id = enemy_id
        self.scale = scale
        self.first_seen = first_seen


class EnemyDictionary:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.dictionary = self.__get_dictionary()

    def __get_dictionary(self) -> list[EnemyDictionaryItem] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        csv_data = gdg.download("DataLocal", "enemy_dictionary_list.csv")
        if csv_data is None:
            return None

        csv = core.CSV(csv_data)
        data: list[EnemyDictionaryItem] = []

        for row in csv:
            first_seen = None
            if len(row) >= 3:
                first_seen = row[2].to_int()
            data.append(
                EnemyDictionaryItem(row[0].to_int(), row[1].to_int(), first_seen)
            )

        return data

    def get_valid_enemies(self) -> list[int] | None:
        if self.dictionary is None:
            return None

        return [enemy.enemy_id for enemy in self.dictionary]

    def get_invalid_enemies(self, total_enemies: int) -> list[int] | None:
        valid_enemies = self.get_valid_enemies()
        if valid_enemies is None:
            return None

        valid_enemies = set(valid_enemies)

        return list(filter(lambda i: i not in valid_enemies, range(total_enemies)))


class EnemyDescription:
    def __init__(self, trait_str: str, description: list[str] | None):
        self.trait_str = trait_str
        self.description = description


class EnemyDescriptions:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.descriptions = self.__get_descriptions()

    def __get_descriptions(self) -> list[EnemyDescription] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download(
            "resLocal",
            f"EnemyPictureBook_{core.core_data.get_lang(self.save_file)}.csv",
        )
        if data is None:
            return None

        csv = core.CSV(data, core.Delimeter.from_country_code_res(self.save_file.cc))
        descriptions: list[EnemyDescription] = []

        for i, row in enumerate(csv):
            if len(row) == 1:
                descriptions.append(EnemyDescription(row[0].to_str(), None))
            else:
                descriptions.append(
                    EnemyDescription(row[0].to_str(), row[1:].to_str_list())
                )

        return descriptions


class EnemyNames:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.names = self.get_names()

    def get_names(self) -> list[str] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "Enemyname.tsv")
        if data is None:
            return None
        csv = core.CSV(
            data,
            "\t",
            remove_empty=False,
        )
        names: list[str] = []
        for row in csv:
            names.append(row[0].to_str())

        return names

    def get_name(self, id: int) -> str | None:
        if self.names is None:
            return None
        try:
            name = self.names[id]
            if not name:
                return core.core_data.local_manager.get_key(
                    "enemy_not_in_name_list", id=id
                )
        except IndexError:
            return core.core_data.local_manager.get_key("enemy_unknown_name", id=id)
        return name



from __future__ import annotations
from typing import Callable, Any, Iterable
import threading


class Thread:
    def __init__(
        self,
        name: str,
        target: Callable[..., Any],
        args: Iterable[Any] | None = None,
    ):
        self.name = name
        self.target = target
        self.args: Iterable[Any] = args if args is not None else []
        self._thread: threading.Thread | None = None

    def start(self):
        self._thread = threading.Thread(
            target=self.target, args=self.args, name=self.name
        )
        self._thread.start()

    def join(self):
        if self._thread is not None:
            self._thread.join()

    def is_alive(self) -> bool:
        if self._thread is not None:
            return self._thread.is_alive()
        return False

    @staticmethod
    def run(name: str, target: Callable[..., None], args: Any):
        thread = Thread(name, target, args)
        thread.start()
        return thread


def thread_run_many_helper(funcs: list[Callable[..., Any]], *args: list[Any]):
    for i in range(len(funcs)):
        args_ = args[i]
        funcs[i](*args_)
    return


def thread_run_many(
    funcs: list[Callable[..., Any]], args: Any = None, max_threads: int = 16
) -> list[Thread]:
    chunk_size = len(funcs) // max_threads
    if chunk_size == 0:
        chunk_size = 1
    callable_chunks: list[list[Callable[..., Any]]] = []
    args_chunks: list[list[Any]] = []
    for i in range(0, len(funcs), chunk_size):
        callable_chunks.append(funcs[i : i + chunk_size])
        args_chunks.append(args[i : i + chunk_size])

    threads: list[Thread] = []
    for i in range(len(callable_chunks)):
        args_ = args_chunks[i]
        if args is None:
            args_ = []

        threads.append(
            Thread.run(
                "run_many_helper",
                thread_run_many_helper,
                (callable_chunks[i], *args_),
            )
        )

    for thread in threads:
        thread.join()

    return threads



from __future__ import annotations
from bcsfe import core
from bcsfe.cli import color


class Repo:
    def __init__(self, url: str, output_error: bool = True):
        self.url = url
        self.output_error = output_error
        self.success = self.clone()

    def get_repo_name(self) -> str:
        return self.url.split("/")[-1]

    def get_path(self) -> core.Path:
        path = GitHandler.get_repo_folder().add(self.get_repo_name())
        path.generate_dirs()
        return path

    def run_cmd(self, cmd: str) -> bool:
        result = core.Command(cmd).run()
        success = result.exit_code == 0
        if not success and self.output_error:
            color.ColoredText.localize("failed_to_run_git_cmd", cmd=cmd)
        return success

    def clone_to_temp(self, path: core.Path) -> bool:
        cmd = f"git clone {self.url} {path}"
        return self.run_cmd(cmd)

    def clone(self) -> bool:
        if self.is_cloned():
            return True
        cmd = f"git clone {self.url} {self.get_path()}"
        success = self.run_cmd(cmd)
        if not success:
            self.get_path().remove()
        return success

    def pull(self) -> bool:
        cmd = f"git -C {self.get_path()} pull"
        return self.run_cmd(cmd)

    def fetch(self) -> bool:
        cmd = f"git -C {self.get_path()} fetch"
        return self.run_cmd(cmd)

    def get_file(self, file_path: core.Path) -> core.Data | None:
        path = self.get_path().add(file_path)
        try:
            return path.read()
        except FileNotFoundError:
            return None

    def get_temp_file(
        self, temp_folder: core.Path, file_path: core.Path
    ) -> core.Data:
        path = temp_folder.add(file_path)
        return path.read()

    def get_folder(self, folder_path: core.Path) -> core.Path | None:
        path = self.get_path().add(folder_path)
        if path.exists():
            return path
        return None

    def is_cloned(self) -> bool:
        return (
            len(self.get_path().get_dirs()) > 0
            or len(self.get_path().get_paths_dir()) > 0
        )


class GitHandler:
    @staticmethod
    def get_repo_folder() -> core.Path:
        repo_folder = core.Path.get_documents_folder().add("repos")
        repo_folder.generate_dirs()
        return repo_folder

    def get_repo(self, repo_url: str, output_error: bool = True) -> Repo | None:
        repo = Repo(repo_url)
        if repo.success:
            return repo
        if output_error:
            color.ColoredText.localize("failed_to_get_repo", url=repo_url)
        return None

    @staticmethod
    def is_git_installed() -> bool:
        cmd = "git --version"
        return core.Command(cmd).run().exit_code == 0



from __future__ import annotations

import datetime
from math import inf, isnan
import math
from typing import Any
from bcsfe import core
from bcsfe.cli import dialog_creator, color


class EndlessItem:
    def __init__(
        self, active: bool, unknown: bool, amount: int, start: float, end: float
    ):
        self.active = active
        self.unknown = unknown
        self.amount = amount
        self.start = start
        self.end = end

    @staticmethod
    def init() -> EndlessItem:
        return EndlessItem(False, False, 0, 0, 0)

    @staticmethod
    def read(stream: core.Data) -> EndlessItem:
        return EndlessItem(
            stream.read_bool(),
            stream.read_bool(),
            stream.read_byte(),
            stream.read_double(),
            stream.read_double(),
        )

    def write(self, stream: core.Data):
        stream.write_bool(self.active)
        stream.write_bool(self.unknown)
        stream.write_byte(self.amount)
        stream.write_double(self.start)
        stream.write_double(self.end)

    def serialize(self) -> dict[str, Any]:
        return {
            "active": self.active,
            "unknown": self.unknown,
            "amount": self.amount,
            "start": self.start,
            "end": self.end,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> EndlessItem:
        return EndlessItem(
            data.get("active", False),
            data.get("unknown", False),
            data.get("amount", 0),
            data.get("start", 0.0),
            data.get("end", 0.0),
        )

    def get_endless_duration(self) -> datetime.timedelta | None:
        if not self.active:
            return datetime.timedelta()

        if self.end == inf:
            return None
        if math.isnan(self.end) or math.isnan(self.start):
            return None

        return datetime.timedelta(
            seconds=self.end - self.start + (self.amount * 3 * 60 * 60)
        )

    def get_endless_duration_formatted(self) -> str:
        duration = self.get_endless_duration()

        if duration is None:
            return core.localize("infinity_duration")

        days = duration.days
        hours, rem = divmod(duration.seconds, 3600)
        minutes, seconds = divmod(rem, 60)

        return core.localize(
            "duration", days=days, hours=hours, minutes=minutes, seconds=seconds
        )

    def set_duration_mins(self, mins: float, amount: int):
        self.active = True
        self.unknown = True
        self.amount = amount
        self.start = datetime.datetime.now(datetime.timezone.utc).timestamp()
        self.end = self.start + mins * 60


class BattleItem:
    def __init__(self, amount: int):
        self.amount = amount
        self.locked = False

        self.endless_item = EndlessItem.init()

    @staticmethod
    def init() -> BattleItem:
        return BattleItem(0)

    @staticmethod
    def read_amount(stream: core.Data) -> BattleItem:
        return BattleItem(stream.read_int())

    def write_amount(self, stream: core.Data):
        stream.write_int(self.amount)

    def read_locked(self, stream: core.Data):
        self.locked = stream.read_bool()

    def write_locked(self, stream: core.Data):
        stream.write_bool(self.locked)

    def read_endless_items(self, stream: core.Data):
        self.endless_item = EndlessItem.read(stream)

    def write_endless_items(self, stream: core.Data):
        self.endless_item.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "amount": self.amount,
            "locked": self.locked,
            "endless": self.endless_item.serialize(),
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> BattleItem:
        battle_item = BattleItem(data.get("amount", 0))
        battle_item.locked = data.get("locked", False)
        battle_item.endless_item = EndlessItem.deserialize(data.get("endless", {}))
        return battle_item

    def __repr__(self):
        try:
            return f"BattleItem({self.amount}, {self.locked}, {self.endless_item})"
        except AttributeError:
            return f"BattleItem({self.amount}, {self.endless_item})"

    def __str__(self):
        return self.__repr__()


class BattleItems:
    def __init__(self, items: list[BattleItem]):
        self.items = items
        self.lock_item = False

    @staticmethod
    def init() -> BattleItems:
        return BattleItems([BattleItem.init() for _ in range(6)])

    @staticmethod
    def read_items(stream: core.Data) -> BattleItems:
        total_items = 6
        items = [BattleItem.read_amount(stream) for _ in range(total_items)]
        return BattleItems(items)

    def write_items(self, stream: core.Data):
        for item in self.items:
            item.write_amount(stream)

    def read_locked_items(self, stream: core.Data):
        self.lock_item = stream.read_bool()
        for item in self.items:
            item.read_locked(stream)

    def write_locked_items(self, stream: core.Data):
        stream.write_bool(self.lock_item)
        for item in self.items:
            item.write_locked(stream)

    def read_endless_items(self, stream: core.Data):
        for i in range(6):
            if i >= len(self.items):
                _ = EndlessItem.read(stream)  # ensure we still read 6 items
            else:
                item = self.items[i]
                item.read_endless_items(stream)

    def write_endless_items(self, stream: core.Data):
        for i in range(6):
            if i >= len(self.items):
                EndlessItem.init().write(stream)  # ensure we still write 6 items
            else:
                item = self.items[i]
                item.write_endless_items(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "items": [item.serialize() for item in self.items],
            "lock_item": self.lock_item,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> BattleItems:
        battle_items = BattleItems(
            [BattleItem.deserialize(item) for item in data.get("items", [])]
        )
        battle_items.lock_item = data.get("lock_item", False)
        return battle_items

    def __repr__(self):
        return f"BattleItems({self.items})"

    def __str__(self):
        return f"BattleItems({self.items})"

    def get_names(self, save_file: core.SaveFile) -> list[str] | None:
        names = core.core_data.get_gatya_item_names(save_file).names
        if names is None:
            return None
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(3)
        if items is None:
            return None

        names = [names[item.id] for item in items]
        return names

    def edit(self, save_file: core.SaveFile):
        group_name = save_file.get_localizable().get("shop_category1")
        if group_name is None:
            group_name = core.core_data.local_manager.get_key("battle_items")
        item_names = self.get_names(save_file)
        if item_names is None:
            return
        current_values = [item.amount for item in self.items]
        values = dialog_creator.MultiEditor.from_reduced(
            group_name,
            item_names,
            current_values,
            core.core_data.max_value_manager.get("battle_items"),
        ).edit()
        for i, value in enumerate(values):
            self.items[i].amount = value

    def edit_endless_items(self, save_file: core.SaveFile):
        item_names = self.get_names(save_file)
        if item_names is None:
            return

        current_values = [
            item.endless_item.get_endless_duration_formatted() for item in self.items
        ]

        (options, all_at_once) = dialog_creator.ChoiceInput.from_reduced(
            [core.localize("endless_item_item", item=item) for item in item_names],
            current_values,
            localize_options=False,
            dialog="select_option",
        ).multiple_choice(False)

        if options is None:
            return

        infinity_str = core.localize("infinity")

        if all_at_once:
            val = dialog_creator.StringInput().get_input_locale_while(
                "enter_duration_minutes", {}
            )
            if val is None:
                return

            if val.lower() == infinity_str.lower():
                val = inf
            else:
                try:
                    val = float(val)
                except ValueError:
                    return

            for item in self.items:
                item.endless_item.set_duration_mins(val, 0)
        else:
            for opt in options:
                val = dialog_creator.StringInput().get_input_locale_while(
                    "enter_duration_minutes_item", {"item": item_names[opt]}
                )
                if val is None:
                    return

                if val.lower() == infinity_str.lower():
                    val = inf
                else:
                    try:
                        val = float(val)
                    except ValueError:
                        color.ColoredText.localize("invalid_minute_count")
                        continue

                self.items[opt].endless_item.set_duration_mins(val, 0)

        color.ColoredText.localize("endless_items_success")



from __future__ import annotations

from bcsfe import core
from bcsfe.cli import color
from bcsfe.core import io
from bcsfe.core.io.command import CommandResult


class WayDroidNotInstalledError(Exception):
    def __init__(self, result: CommandResult):
        self.result = result


class WayDroidHandler(io.root_handler.RootHandler):
    def __init__(self):
        self.check_waydroid_installed()

        self.adb_handler = io.adb_handler.AdbHandler(root=False)

        self.package_name = None

    def set_package_name(self, package_name: str):
        self.package_name = package_name
        self.adb_handler.set_package_name(self.package_name)

    @staticmethod
    def display_waydroid_not_installed(e: WayDroidNotInstalledError):
        color.ColoredText.localize("waydroid_not_installed", error=e)
        return

    @staticmethod
    def check_waydroid_installed():
        result = io.command.Command("waydroid -V").run()
        if not result.success:
            raise WayDroidNotInstalledError(result)

    def run_shell_cmd(self, command: str) -> core.CommandResult:
        cmd = "waydroid shell"
        use_pkexec = core.core_data.config.get_bool(core.ConfigKey.USE_PKEXEC_WAYDROID)
        if use_pkexec:
            cmd = "pkexec " + cmd
        return io.command.Command(cmd).run(f"{command}")

    def pull_file(
        self, device_path: core.Path, local_path: core.Path
    ) -> core.CommandResult:
        # copy file to sdcard

        result = self.run_shell_cmd(
            f"cp {device_path.to_str_forwards()} /sdcard/{device_path.basename()} && chmod o+rw /sdcard/{device_path.basename()}"
        )

        if not result.success:
            return result

        device_path = core.Path("/sdcard/").add(device_path.basename())

        # adb pull

        result = self.adb_handler.adb_pull_file(device_path, local_path)
        if not result.success:
            return result

        # delete /sdcard file again
        #
        return self.adb_handler.run_shell(f"rm /sdcard/{device_path.basename()}")

    def push_file(
        self, local_path: core.Path, device_path: core.Path
    ) -> core.CommandResult:
        original_device_path = device_path.copy_object()

        device_path = core.Path("/sdcard/").add(device_path.basename())

        # push to /sdcard with adb

        import time

        time.sleep(0.25)
        result = self.adb_handler.adb_push_file(local_path, device_path)

        if not result.success:
            return result

        result = self.run_shell_cmd(
            f"cp '/sdcard/{device_path.basename()}' '{original_device_path.to_str_forwards()}' && chmod o+rw '{original_device_path.to_str_forwards()}'"
        )

        if not result.success:
            return result

        # remove temp file
        #
        return self.adb_handler.run_shell(f"rm '/sdcard/{device_path.basename()}'")

    def get_battlecats_packages(self) -> list[str]:
        cmd = "find /data/data/ -name SAVE_DATA -mindepth 3 -maxdepth 3"
        result = self.run_shell_cmd(cmd)

        if not result.success:
            return []

        packages: list[str] = []

        for package in result.result.split("\n"):
            parts = package.split("/")
            if len(parts) < 4:
                continue

            packages.append(package.split("/")[3])

        return packages

    def save_battlecats_save(self, local_path: core.Path) -> core.CommandResult:
        return self.pull_file(self.get_battlecats_save_path(), local_path)

    def load_battlecats_save(self, local_path: core.Path) -> core.CommandResult:
        return self.push_file(local_path, self.get_battlecats_save_path())

    def run_game(self) -> core.CommandResult:
        return self.adb_handler.run_game()

    def close_game(self) -> core.CommandResult:
        return self.adb_handler.close_game()



from __future__ import annotations
import enum
import hashlib
import hmac
import random
from bcsfe import core


class HashAlgorithm(enum.Enum):
    """An enum representing a hash algorithm."""

    MD5 = enum.auto()
    SHA1 = enum.auto()
    SHA256 = enum.auto()


class Hash:
    """A class to hash data."""

    def __init__(self, algorithm: HashAlgorithm):
        """Initializes a new instance of the Hash class.

        Args:
            algorithm (HashAlgorithm): The hash algorithm to use.
        """
        self.algorithm = algorithm

    def get_hash(
        self,
        data: core.Data,
        length: int | None = None,
    ) -> core.Data:
        """Gets the hash of the given data.

        Args:
            data (core.Data): The data to hash.
            length (int | None, optional): The length of the hash. Defaults to None.

        Raises:
            ValueError: Invalid hash algorithm.

        Returns:
            core.Data: The hash of the data.
        """
        if self.algorithm == HashAlgorithm.MD5:
            hash = hashlib.md5()
        elif self.algorithm == HashAlgorithm.SHA1:
            hash = hashlib.sha1()
        elif self.algorithm == HashAlgorithm.SHA256:
            hash = hashlib.sha256()
        else:
            raise ValueError("Invalid hash algorithm")
        hash.update(data.get_bytes())
        if length is None:
            return core.Data(hash.digest())
        return core.Data(hash.digest()[:length])


class Random:
    """A class to get random data"""

    @staticmethod
    def get_bytes(length: int) -> bytes:
        """Gets random bytes.

        Args:
            length (int): The length of the bytes.

        Returns:
            bytes: The random bytes.
        """
        return bytes(random.getrandbits(8) for _ in range(length))

    @staticmethod
    def get_alpha_string(length: int) -> str:
        """Gets a random string of the given length.

        Args:
            length (int): The length of the string.

        Returns:
            str: The random string.
        """
        characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        return "".join(random.choice(characters) for _ in range(length))

    @staticmethod
    def get_hex_string(length: int) -> str:
        """Gets a random hex string of the given length.

        Args:
            length (int): The length of the string.

        Returns:
            str: The random string.
        """
        characters = "0123456789abcdef"
        return "".join(random.choice(characters) for _ in range(length))

    @staticmethod
    def get_digits_string(length: int) -> str:
        """Gets a random digits string of the given length.

        Args:
            length (int): The length of the string.

        Returns:
            str: The random string.
        """
        characters = "0123456789"
        return "".join(random.choice(characters) for _ in range(length))


class Hmac:
    def __init__(self, algorithm: HashAlgorithm):
        self.algorithm = algorithm

    def get_hmac(self, key: core.Data, data: core.Data) -> core.Data:
        if self.algorithm == HashAlgorithm.MD5:
            alg = hashlib.md5
        elif self.algorithm == HashAlgorithm.SHA1:
            alg = hashlib.sha1
        elif self.algorithm == HashAlgorithm.SHA256:
            alg = hashlib.sha256
        else:
            raise ValueError("Invalid hash algorithm")
        hmac_data = hmac.new(
            key.get_bytes(), data.get_bytes(), digestmod=alg
        ).digest()
        return core.Data(hmac_data)


class NyankoSignature:
    def __init__(self, inquiry_code: str, data: str):
        self.inquiry_code = inquiry_code
        self.data = data

    def generate_signature(self) -> str:
        """Generates a signature from the inquiry code and data.

        Returns:
            str: The signature.
        """
        random_data = Random.get_hex_string(64)
        key = self.inquiry_code + random_data
        hmac_ = Hmac(HashAlgorithm.SHA256)
        signature = hmac_.get_hmac(core.Data(key), core.Data(self.data))

        return random_data + signature.to_hex()

    def generate_signature_v1(self) -> str:
        """Generates a signature from the inquiry code and data.

        Returns:
            str: The signature.
        """

        data = self.data + self.data  # repeat data for some reason
        random_data = Random.get_hex_string(40)
        key = self.inquiry_code + random_data
        hmac_ = Hmac(HashAlgorithm.SHA1)
        signature = hmac_.get_hmac(core.Data(key), core.Data(data))

        return random_data + signature.to_hex()



from __future__ import annotations
import random
from bcsfe import core
from bcsfe.cli import dialog_creator, color, edits
from bcsfe.core.game.catbase.gatya_item import GatyaItemCategory


class BasicItems:
    @staticmethod
    def get_name(name: str | None, key: str) -> str:
        if name is None:
            return core.core_data.local_manager.get_key(key)
        return name.strip()

    @staticmethod
    def reset_golden_cat_cpus(save_file: core.SaveFile):
        save_file.golden_cpu_count = 0

        color.ColoredText.localize("reset_golden_cat_cpus_success")

    @staticmethod
    def edit_catfood(save_file: core.SaveFile):
        should_exit = not dialog_creator.YesNoInput().get_input_once("catfood_warning")
        if should_exit:
            return

        name = core.core_data.get_gatya_item_names(save_file).get_name(22)
        original_amount = save_file.catfood
        save_file.catfood = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "catfood"),
            save_file.catfood,
            core.core_data.max_value_manager.get("catfood"),
        ).edit()
        change = save_file.catfood - original_amount
        core.BackupMetaData(save_file).add_managed_item(
            core.ManagedItem.from_change(change, core.ManagedItemType.CATFOOD)
        )

    @staticmethod
    def edit_xp(save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(6)
        save_file.xp = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "xp"),
            save_file.xp,
            core.core_data.max_value_manager.get("xp"),
        ).edit()

    @staticmethod
    def edit_normal_tickets(save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(20)
        save_file.normal_tickets = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "normal_tickets"),
            save_file.normal_tickets,
            core.core_data.max_value_manager.get("normal_tickets"),
        ).edit()

    @staticmethod
    def edit_100_million_ticket(save_file: core.SaveFile):
        color.ColoredText.localize("100_million_warn")
        name = core.core_data.get_gatya_item_names(save_file).get_name(212)
        save_file.hundred_million_ticket = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "100_million_tickets"),
            save_file.hundred_million_ticket,
            core.core_data.max_value_manager.get("100_million_tickets"),
        ).edit()

    @staticmethod
    def get_bannable_feature_options(feature_name: str, safe_feature_name: str) -> int:
        feature_name = core.core_data.local_manager.get_key(feature_name)
        safe_feature_name = core.core_data.local_manager.get_key(safe_feature_name)

        options = [
            core.core_data.local_manager.get_key(
                "continue_editing", feature_name=feature_name
            ),
            core.core_data.local_manager.get_key(
                "go_to_safe_feature", safer_feature_name=safe_feature_name
            ),
            core.core_data.local_manager.get_key(
                "cancel_editing", feature_name=feature_name
            ),
        ]
        option = dialog_creator.ChoiceInput(
            options,
            options,
            [],
            {"feature_name": feature_name},
            "select_an_option_to_continue",
        ).single_choice()
        if option is None:
            return 2
        option -= 1
        return option

    @staticmethod
    def edit_rare_tickets(save_file: core.SaveFile):
        color.ColoredText.localize("rare_ticket_warning")
        name = core.core_data.get_gatya_item_names(save_file).get_name(21)
        option = BasicItems.get_bannable_feature_options(
            "rare_tickets_l", "rare_ticket_trade_l"
        )
        if option == 2:
            return
        if option == 1:
            return edits.rare_ticket_trade.RareTicketTrade.rare_ticket_trade(save_file)

        original_amount = save_file.rare_tickets
        save_file.rare_tickets = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "rare_tickets"),
            save_file.rare_tickets,
            core.core_data.max_value_manager.get("rare_tickets"),
        ).edit()
        change = save_file.rare_tickets - original_amount
        core.BackupMetaData(save_file).add_managed_item(
            core.ManagedItem.from_change(change, core.ManagedItemType.RARE_TICKET)
        )

    @staticmethod
    def edit_platinum_tickets(save_file: core.SaveFile):
        color.ColoredText.localize("platinum_ticket_warning")
        name = core.core_data.get_gatya_item_names(save_file).get_name(29)
        option = BasicItems.get_bannable_feature_options(
            "platinum_tickets_l", "platinum_shards_l"
        )
        if option == 2:
            return
        if option == 1:
            return edits.basic_items.BasicItems.edit_platinum_shards(save_file)

        original_amount = save_file.platinum_tickets
        save_file.platinum_tickets = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "platinum_tickets"),
            save_file.platinum_tickets,
            core.core_data.max_value_manager.get("platinum_tickets"),
        ).edit()
        change = save_file.platinum_tickets - original_amount
        core.BackupMetaData(save_file).add_managed_item(
            core.ManagedItem.from_change(change, core.ManagedItemType.PLATINUM_TICKET)
        )

    @staticmethod
    def edit_legend_tickets(save_file: core.SaveFile):
        should_exit = not dialog_creator.YesNoInput().get_input_once(
            "legend_ticket_warning"
        )
        if should_exit:
            return
        name = core.core_data.get_gatya_item_names(save_file).get_name(145)
        original_amount = save_file.legend_tickets
        save_file.legend_tickets = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "legend_tickets"),
            save_file.legend_tickets,
            core.core_data.max_value_manager.get("legend_tickets"),
        ).edit()
        change = save_file.legend_tickets - original_amount
        core.BackupMetaData(save_file).add_managed_item(
            core.ManagedItem.from_change(change, core.ManagedItemType.LEGEND_TICKET)
        )

    @staticmethod
    def edit_platinum_shards(save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(157)
        platinum_ticket_amount = save_file.platinum_tickets
        max_value = (
            core.core_data.max_value_manager.get("platinum_tickets")
            - platinum_ticket_amount
        ) * 10 + 9

        max_value = max(0, max_value)
        save_file.platinum_shards = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "platinum_shards"),
            save_file.platinum_shards,
            max_value,
        ).edit()

    @staticmethod
    def edit_np(save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(7)
        save_file.np = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "np"),
            save_file.np,
            core.core_data.max_value_manager.get("np"),
        ).edit()

    @staticmethod
    def edit_leadership(save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(105)
        save_file.leadership = dialog_creator.SingleEditor(
            BasicItems.get_name(name, "leadership"),
            save_file.leadership,
            core.core_data.max_value_manager.get("leadership"),
        ).edit()

    @staticmethod
    def edit_battle_items(save_file: core.SaveFile):
        save_file.battle_items.edit(save_file)

    @staticmethod
    def edit_battle_items_endless(save_file: core.SaveFile):
        save_file.battle_items.edit_endless_items(save_file)

    @staticmethod
    def edit_catamins(save_file: core.SaveFile):
        names_o = core.core_data.get_gatya_item_names(save_file)
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(6)
        if items is None:
            return
        names: list[str] = []
        for item in items:
            name = names_o.get_name(item.id)
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_catamin_name", id=item.id
                )
            names.append(name)
        values = dialog_creator.MultiEditor.from_reduced(
            "catamins",
            names,
            save_file.catamins,
            core.core_data.max_value_manager.get("catamins"),
            group_name_localized=True,
        ).edit()
        save_file.catamins = values

    @staticmethod
    def edit_catseyes(save_file: core.SaveFile):
        names_o = core.core_data.get_gatya_item_names(save_file)
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(5)
        if items is None:
            return
        names: list[str] = []
        for item in items:
            name = names_o.get_name(item.id)
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_catseye_name", id=item.id
                )
            names.append(name)

        values = dialog_creator.MultiEditor.from_reduced(
            "catseyes",
            names,
            save_file.catseyes,
            core.core_data.max_value_manager.get("catseyes"),
            group_name_localized=True,
        ).edit()
        save_file.catseyes = values

    @staticmethod
    def edit_treasure_chests(save_file: core.SaveFile):
        names_o = core.core_data.get_gatya_item_names(save_file)
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(
            GatyaItemCategory.TREASURE_CHESTS
        )
        if items is None:
            return
        names: list[str] = []
        for item in items[: len(save_file.treasure_chests)]:
            name = names_o.get_name(item.id)
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_treasure_chest_name", id=item.id
                )
            names.append(name)

        values = dialog_creator.MultiEditor.from_reduced(
            "treasure_chests",
            names,
            save_file.treasure_chests,
            core.core_data.max_value_manager.get("treasure_chests"),
            group_name_localized=True,
        ).edit()
        save_file.treasure_chests = values

    @staticmethod
    def edit_catfruit(save_file: core.SaveFile):
        names = core.Matatabi(save_file).get_names()
        if names is None:
            return
        new_names: list[str] = []
        for i, name in enumerate(names):
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_catfruit_name", id=i
                )
            new_names.append(name)
        names = new_names

        extra = len(save_file.catfruit) - len(names)
        if extra > 0:
            for i in range(extra):
                names.append(
                    core.core_data.local_manager.get_key(
                        "unknown_catfruit_name", id=i + len(names)
                    )
                )

        if save_file.game_version < 110400:
            max_value = core.core_data.max_value_manager.get_old("catfruit")
            cumulative_max = True
        else:
            max_value = core.core_data.max_value_manager.get_new("catfruit")
            cumulative_max = False

        names = names[: len(save_file.catfruit)]

        values = dialog_creator.MultiEditor.from_reduced(
            "catfruit",
            names,
            save_file.catfruit,
            max_value,
            group_name_localized=True,
            cumulative_max=cumulative_max,
        ).edit()
        save_file.catfruit = values

    @staticmethod
    def set_restart_pack(save_file: core.SaveFile):
        save_file.restart_pack = 1
        name = core.core_data.get_gatya_item_names(save_file).get_name(123)
        color.ColoredText.localize("value_gave", name=name)

    @staticmethod
    def edit_inquiry_code(save_file: core.SaveFile):
        should_exit = not dialog_creator.YesNoInput().get_input_once(
            "inquiry_code_warning"
        )
        if should_exit:
            return
        item_name = save_file.get_localizable().get("autoSave_txt5")
        save_file.inquiry_code = dialog_creator.StringEditor(
            BasicItems.get_name(item_name, "inquiry_code"),
            save_file.inquiry_code,
        ).edit()

    @staticmethod
    def edit_password_refresh_token(save_file: core.SaveFile):
        should_exit = not dialog_creator.YesNoInput().get_input_once(
            "password_refresh_token_warning"
        )
        if should_exit:
            return
        save_file.password_refresh_token = dialog_creator.StringEditor(
            "password_refresh_token",
            save_file.password_refresh_token,
            item_localized=True,
        ).edit()

    @staticmethod
    def edit_scheme_items(save_file: core.SaveFile):
        save_file.scheme_items.edit(save_file)

    @staticmethod
    def edit_engineers(save_file: core.SaveFile):
        save_file.ototo.edit_engineers(save_file)

    @staticmethod
    def edit_base_materials(save_file: core.SaveFile):
        save_file.ototo.base_materials.edit_base_materials(save_file)

    @staticmethod
    def edit_rare_gatya_seed(save_file: core.SaveFile):
        save_file.gatya.edit_rare_gatya_seed()

    @staticmethod
    def edit_normal_gatya_seed(save_file: core.SaveFile):
        save_file.gatya.edit_normal_gatya_seed()

    @staticmethod
    def edit_event_gatya_seed(save_file: core.SaveFile):
        save_file.gatya.edit_event_gatya_seed()

    @staticmethod
    def edit_unlocked_slots(save_file: core.SaveFile):
        save_file.lineups.edit_unlocked_slots()

    @staticmethod
    def edit_labyrinth_medals(save_file: core.SaveFile):
        names_o = core.core_data.get_gatya_item_names(save_file)
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(11)
        if items is None:
            return
        names: list[str] = []
        for item in items:
            name = names_o.get_name(item.id)
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_labyrinth_medal_name", id=item.id
                )
            names.append(name)

        values = dialog_creator.MultiEditor.from_reduced(
            "labyrinth_medals",
            names,
            save_file.labyrinth_medals,
            core.core_data.max_value_manager.get("labyrinth_medals"),
            group_name_localized=True,
        ).edit()
        save_file.labyrinth_medals = values

    @staticmethod
    def edit_special_skills(save_file: core.SaveFile):
        save_file.special_skills.edit(save_file)

    @staticmethod
    def unlock_equip_menu(save_file: core.SaveFile):
        save_file.unlock_equip_menu()
        color.ColoredText.localize("equip_menu_unlocked")

    @staticmethod
    def allow_filibuster_stage_reclearing(save_file: core.SaveFile):
        save_file.filibuster_stage_enabled = True
        save_file.filibuster_stage_id = random.randint(0, 47)
        color.ColoredText.localize("filibuster_stage_reclearing_allowed")



from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


@dataclass
class MemberName:
    member_id: int
    rarity: int
    bonus: int
    name: str
    rarity_name: str
    description: list[str]


class GamatotoMembersName:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.members = self.read_members()

    def read_members(self) -> list[MemberName] | None:
        members: list[MemberName] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download(
            "resLocal",
            f"GamatotoExpedition_Members_name_{core.core_data.get_lang(self.save_file)}.csv",
        )
        if data is None:
            return None
        csv = core.CSV(
            data,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
            remove_empty=False,
        )
        for line in csv.lines[1:]:
            if line[0].to_int() == -1:
                continue
            members.append(
                MemberName(
                    line[0].to_int(),
                    line[1].to_int(),
                    line[2].to_int(),
                    line[3].to_str(),
                    line[4].to_str(),
                    line[5:].to_str_list(),
                )
            )
        return members

    def get_member(self, member_id: int) -> MemberName | None:
        if self.members is None:
            return None
        for member in self.members:
            if member.member_id == member_id:
                return member
        return None

    def get_members_from_ids(self, ids: list[int]) -> list[MemberName | None]:
        return [self.get_member(id) for id in ids]

    def get_all_rarity(self, rarity: int) -> list[MemberName] | None:
        if self.members is None:
            return None

        return [member for member in self.members if member.rarity == rarity]

    def get_members_from_helpers(
        self, helpers: Helpers
    ) -> list[MemberName | None]:
        return self.get_members_from_ids(
            [helper.id for helper in helpers.helpers if helper.is_valid()]
        )

    def get_all_rarity_names(self) -> list[str] | None:
        if self.members is None:
            return None
        names: dict[int, str] = {}
        for member in self.members:
            names[member.rarity] = member.rarity_name
        return [names[i] for i in range(len(names))]


@dataclass
class GamatotoLevel:
    level: int
    xp_needed: int
    discovery_bonus: int
    skin: int


@dataclass
class GamatotoLimit:
    max_level: int
    total_stages: int
    total_helpers: int


class GamatotoLevels:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.levels = self.read_levels()
        self.limit = self.read_max_level()

    def read_levels(self) -> list[GamatotoLevel] | None:
        levels: list[GamatotoLevel] = []
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "GamatotoExpedition.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        for i, line in enumerate(csv):
            levels.append(
                GamatotoLevel(
                    i + 1, line[0].to_int(), line[1].to_int(), line[2].to_int()
                )
            )
        return levels

    def read_max_level(self) -> GamatotoLimit | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "GamatotoExpedition_Limit.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        line = csv[0]
        return GamatotoLimit(
            line[0].to_int(), line[1].to_int(), line[2].to_int()
        )

    def get_level(self, level: int) -> GamatotoLevel | None:
        if self.levels is None:
            return None
        if level < 1:
            return None
        return self.levels[level - 1]

    def get_all_levels(self) -> list[GamatotoLevel] | None:
        return self.levels

    def get_level_from_xp(self, xp: int) -> GamatotoLevel | None:
        if self.levels is None or self.limit is None:
            return None
        for level in self.levels:
            if level.level >= self.limit.max_level:
                break
            if level.xp_needed == -1:
                continue
            if xp < level.xp_needed:
                return level
        if self.limit.max_level >= len(self.levels):
            return self.levels[-1]
        return self.levels[self.limit.max_level - 1]

    def get_xp_from_level(self, level: int) -> int | None:
        if self.levels is None:
            return None
        level -= 1
        if level < 1:
            return 0
        return self.levels[level - 1].xp_needed

    def get_max_level(self) -> int | None:
        if self.limit is None:
            return None
        return self.limit.max_level

    def get_total_stages(self) -> int | None:
        if self.limit is None:
            return None
        return self.limit.total_stages

    def get_total_helpers(self) -> int | None:
        if self.limit is None:
            return None
        return self.limit.total_helpers


class Helper:
    def __init__(self, id: int):
        self.id = id

    @staticmethod
    def init() -> Helper:
        return Helper(-1)

    @staticmethod
    def read(stream: core.Data) -> Helper:
        id = stream.read_int()
        return Helper(id)

    def write(self, stream: core.Data):
        stream.write_int(self.id)

    def serialize(self) -> int:
        return self.id

    @staticmethod
    def deserialize(data: int) -> Helper:
        return Helper(data)

    def __repr__(self) -> str:
        return f"Helper(id={self.id!r})"

    def __str__(self) -> str:
        return f"Helper(id={self.id!r})"

    def is_valid(self) -> bool:
        return self.id != -1


class Helpers:
    def __init__(self, helpers: list[Helper]):
        self.helpers = helpers

    @staticmethod
    def init() -> Helpers:
        return Helpers([])

    @staticmethod
    def read(stream: core.Data) -> Helpers:
        total = stream.read_int()
        helpers: list[Helper] = []
        for _ in range(total):
            helpers.append(Helper.read(stream))
        return Helpers(helpers)

    def write(self, stream: core.Data):
        stream.write_int(len(self.helpers))
        for helper in self.helpers:
            helper.write(stream)

    def serialize(self) -> list[int]:
        return [helper.serialize() for helper in self.helpers]

    @staticmethod
    def deserialize(data: list[int]) -> Helpers:
        return Helpers([Helper.deserialize(helper) for helper in data])

    def __repr__(self) -> str:
        return f"Helpers(helpers={self.helpers!r})"

    def __str__(self) -> str:
        return f"Helpers(helpers={self.helpers!r})"


class Gamatoto:
    def __init__(
        self,
        remaining_seconds: float,
        return_flag: bool,
        xp: int,
        dest_id: int,
        recon_length: int,
        unknown: int,
        notif_value: int,
    ):
        self.remaining_seconds = remaining_seconds
        self.return_flag = return_flag
        self.xp = xp
        self.dest_id = dest_id
        self.recon_length = recon_length
        self.unknown = unknown
        self.notif_value = notif_value
        self.helpers = Helpers.init()
        self.is_ad_present = False
        self.skin = 0
        self.collab_flags: dict[int, bool] = {}
        self.collab_durations: dict[int, float] = {}

    @staticmethod
    def init() -> Gamatoto:
        return Gamatoto(
            0.0,
            False,
            0,
            0,
            0,
            0,
            0,
        )

    @staticmethod
    def read(stream: core.Data) -> Gamatoto:
        remaining_seconds = stream.read_double()
        return_flag = stream.read_bool()
        xp = stream.read_int()
        dest_id = stream.read_int()
        recon_length = stream.read_int()
        unknown = stream.read_int()
        notif_value = stream.read_int()
        return Gamatoto(
            remaining_seconds,
            return_flag,
            xp,
            dest_id,
            recon_length,
            unknown,
            notif_value,
        )

    def write(self, stream: core.Data):
        stream.write_double(self.remaining_seconds)
        stream.write_bool(self.return_flag)
        stream.write_int(self.xp)
        stream.write_int(self.dest_id)
        stream.write_int(self.recon_length)
        stream.write_int(self.unknown)
        stream.write_int(self.notif_value)

    def read_2(self, stream: core.Data):
        self.helpers = Helpers.read(stream)
        self.is_ad_present = stream.read_bool()

    def write_2(self, stream: core.Data):
        self.helpers.write(stream)
        stream.write_bool(self.is_ad_present)

    def read_skin(self, stream: core.Data):
        self.skin = stream.read_int()

    def write_skin(self, stream: core.Data):
        stream.write_int(self.skin)

    def read_collab_data(self, stream: core.Data):
        self.collab_flags: dict[int, bool] = stream.read_int_bool_dict()
        self.collab_durations: dict[int, float] = stream.read_int_double_dict()

    def write_collab_data(self, stream: core.Data):
        stream.write_int_bool_dict(self.collab_flags)
        stream.write_int_double_dict(self.collab_durations)

    def serialize(self) -> dict[str, Any]:
        return {
            "remaining_seconds": self.remaining_seconds,
            "return_flag": self.return_flag,
            "xp": self.xp,
            "dest_id": self.dest_id,
            "recon_length": self.recon_length,
            "unknown": self.unknown,
            "notif_value": self.notif_value,
            "helpers": self.helpers.serialize(),
            "is_ad_present": self.is_ad_present,
            "skin": self.skin,
            "collab_flags": self.collab_flags,
            "collab_durations": self.collab_durations,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Gamatoto:
        gamatoto = Gamatoto(
            data.get("remaining_seconds", 0.0),
            data.get("return_flag", False),
            data.get("xp", 0),
            data.get("dest_id", 0),
            data.get("recon_length", 0),
            data.get("unknown", 0),
            data.get("notif_value", 0),
        )
        gamatoto.helpers = Helpers.deserialize(data.get("helpers", []))
        gamatoto.is_ad_present = data.get("is_ad_present", False)
        gamatoto.skin = data.get("skin", 0)
        gamatoto.collab_flags = data.get("collab_flags", {})
        gamatoto.collab_durations = data.get("collab_durations", {})
        return gamatoto

    def __repr__(self):
        return (
            f"Gamatoto(remaining_seconds={self.remaining_seconds!r}, "
            f"return_flag={self.return_flag!r}, xp={self.xp!r}, "
            f"dest_id={self.dest_id!r}, recon_length={self.recon_length!r}, "
            f"unknown={self.unknown!r}, notif_value={self.notif_value!r}, "
            f"helpers={self.helpers!r}, is_ad_present={self.is_ad_present!r}, "
            f"skin={self.skin!r}, collab_flags={self.collab_flags!r}, "
            f"collab_durations={self.collab_durations!r})"
        )

    def __str__(self):
        return self.__repr__()

    def edit_xp(self, save_file: core.SaveFile):
        gamatoto_levels = core.core_data.get_gamatoto_levels(save_file)
        current_level = gamatoto_levels.get_level_from_xp(self.xp)
        if current_level is None:
            return
        xp = self.xp

        color.ColoredText.localize(
            "gamatoto_level_current", level=current_level.level, xp=xp
        )
        choice = dialog_creator.ChoiceInput(
            ["enter_raw_gamatoto_xp", "enter_gamatoto_level"],
            ["enter_raw_gamatoto_xp", "enter_gamatoto_level"],
            [],
            {},
            "edit_gamatoto_level_q",
            single_choice=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1

        if choice == 0:
            xp = dialog_creator.SingleEditor(
                "gamatoto_xp", self.xp, None, localized_item=True
            ).edit()
            current_level = gamatoto_levels.get_level_from_xp(xp)
        elif choice == 1:
            value = dialog_creator.SingleEditor(
                "gamatoto_level",
                current_level.level,
                gamatoto_levels.get_max_level(),
                localized_item=True,
            ).edit()
            xp = gamatoto_levels.get_xp_from_level(value)
            current_level = gamatoto_levels.get_level(value)

        if xp is None:
            return

        self.xp = xp

        if current_level is None:
            return

        color.ColoredText.localize(
            "gamatoto_level_success", level=current_level.level, xp=xp
        )

    def edit_helpers(self, save_file: core.SaveFile):
        members_name = core.core_data.get_gamatoto_members_name(save_file)

        gamatoto_levels = core.core_data.get_gamatoto_levels(save_file)
        max_helpers = gamatoto_levels.get_total_helpers()

        members = members_name.get_members_from_helpers(self.helpers)
        color.ColoredText.localize("current_gamatoto_helpers")
        for member in members:
            if member is None:
                continue
            color.ColoredText.localize(
                "gamatoto_helper",
                name=member.name,
                rarity_name=member.rarity_name,
            )
        rarity_names = members_name.get_all_rarity_names()
        if rarity_names is None:
            return

        total_rarity_amounts: list[int] = [0] * len(rarity_names)
        for helper in self.helpers.helpers:
            if not helper.is_valid():
                continue
            member = members_name.get_member(helper.id)
            if member is None:
                continue
            total_rarity_amounts[member.rarity] += 1

        rarity_amounts = dialog_creator.MultiEditor.from_reduced(
            "gamatoto_helpers",
            rarity_names,
            total_rarity_amounts,
            max_helpers,
            group_name_localized=True,
            cumulative_max=True,
        ).edit()

        helpers: list[Helper] = []
        for i, rarity_amount in enumerate(rarity_amounts):
            rarity_members = members_name.get_all_rarity(i)
            if rarity_members is None:
                continue
            for _ in range(rarity_amount):
                member = rarity_members.pop(0)
                helpers.append(Helper(member.member_id))
        self.helpers = Helpers(helpers)

        members = members_name.get_members_from_helpers(self.helpers)
        color.ColoredText.localize("new_gamatoto_helpers")
        for member in members:
            if member is None:
                continue
            color.ColoredText.localize(
                "gamatoto_helper",
                name=member.name,
                rarity_name=member.rarity_name,
            )


def edit_xp(save_file: core.SaveFile):
    save_file.gamatoto.edit_xp(save_file)


def edit_helpers(save_file: core.SaveFile):
    save_file.gamatoto.edit_helpers(save_file)



from __future__ import annotations
import enum
from typing import Any, Callable

from bcsfe import core
from bcsfe.cli import color, dialog_creator


class SelectMode(enum.Enum):
    AND = 0
    OR = 1
    REPLACE = 2


class CatEditor:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file

    def get_current_cats(self):
        return self.save_file.cats.get_unlocked_cats()

    def get_non_unlocked_cats(self):
        return self.save_file.cats.get_non_unlocked_cats()

    def get_non_gacha_cats(self):
        return self.save_file.cats.get_non_gacha_cats(self.save_file)

    def filter_cats(self, cats: list[core.Cat]) -> list[core.Cat]:
        unlocked_cats = self.get_current_cats()
        return [cat for cat in cats if cat in unlocked_cats]

    def get_cats_rarity(self, rarity: int) -> list[core.Cat]:
        return self.save_file.cats.get_cats_rarity(self.save_file, rarity)

    def get_cats_name(self, name: str) -> list[core.Cat]:
        return self.save_file.cats.get_cats_name(self.save_file, name)

    def get_cats_obtainable(self) -> list[core.Cat] | None:
        return self.save_file.cats.get_cats_obtainable(self.save_file)

    def get_cats_unobtainable(self) -> list[core.Cat] | None:
        return self.save_file.cats.get_cats_non_obtainable(self.save_file)

    def get_cats_gatya_banner(self, gatya_id: int) -> list[core.Cat] | None:
        return self.save_file.cats.get_cats_gatya_banner(self.save_file, gatya_id)

    def print_selected_cats(self, current_cats: list[core.Cat]):
        if len(current_cats) > 50:
            color.ColoredText.localize("total_selected_cats", total=len(current_cats))
        else:
            for cat in current_cats:
                names = cat.get_names_cls(self.save_file)
                if not names:
                    names = [str(cat.id)]
                color.ColoredText.localize("selected_cat", id=cat.id, name=names[0])

    def select(
        self, current_cats: list[core.Cat] | None = None, finish_option: bool = True
    ) -> tuple[list[core.Cat], bool]:
        if current_cats is None:
            current_cats = []
        options: dict[str, Callable[[], Any]] = {
            "select_cats_all": self.save_file.cats.get_all_cats,
            "select_cats_current": self.get_current_cats,
            "select_cats_obtainable": self.get_cats_obtainable,
            "select_cats_id": self.select_id,
            "select_cats_name": self.select_name,
            "select_cats_rarity": self.select_rarity,
            "select_cats_gatya_banner": self.select_gatya_banner,
            "select_cats_not_unlocked": self.get_non_unlocked_cats,
            "select_cats_not_obtainable": self.get_cats_unobtainable,
            "select_cats_non_gatya": self.get_non_gacha_cats,
            "select_cats_game_version": self.select_cats_game_version,
        }
        if finish_option:
            options["finish"] = lambda: None
        option_id = dialog_creator.ChoiceInput(
            list(options), list(options), [], {}, "select_cats", True
        ).single_choice()
        if option_id is None:
            return current_cats, False
        option_id -= 1

        if option_id == len(options) - 1 and finish_option:
            return current_cats, True

        func = options[list(options)[option_id]]
        new_cats = func()

        if new_cats is None:
            return current_cats, False

        if current_cats:
            mode_id = dialog_creator.IntInput().get_basic_input_locale("and_mode_q", {})
            if mode_id is None:
                mode = SelectMode.OR
            elif mode_id == 1:
                mode = SelectMode.AND
            elif mode_id == 2:
                mode = SelectMode.OR
            elif mode_id == 3:
                mode = SelectMode.REPLACE
            else:
                mode = SelectMode.OR
        else:
            mode = SelectMode.OR

        if mode == SelectMode.AND:
            return list(set(current_cats) & set(new_cats)), False
        if mode == SelectMode.OR:
            return list(set(current_cats) | set(new_cats)), False
        if mode == SelectMode.REPLACE:
            return new_cats, False
        return new_cats, False

    def select_id(self) -> list[core.Cat] | None:
        cat_ids = dialog_creator.RangeInput(
            len(self.save_file.cats.cats) - 1
        ).get_input_locale("enter_cat_ids", {})
        if cat_ids is None:
            return None
        return self.save_file.cats.get_cats_by_ids(cat_ids)

    def select_cats_game_version(self) -> list[core.Cat] | None:
        unitbuy = core.UnitBuy(self.save_file)
        if unitbuy.unit_buy is None:
            return None

        versions_set: set[int] = set()
        for cat in unitbuy.unit_buy:
            if cat.game_version == -1:
                continue
            versions_set.add(cat.game_version)

        if not versions_set:
            return None

        versions = list(versions_set)
        versions.sort()

        color.ColoredText.localize("possible_gvs")

        cur_major_v = -1
        for version in versions:
            gv = core.GameVersion(version)
            major_v = gv.get_parts()[0]
            if major_v != cur_major_v:
                if cur_major_v != -1:
                    print()
                cur_major_v = major_v
            else:
                color.ColoredText(", ", end="")
            color.ColoredText(f"<@t>{gv.format()}</>", end="")

        print()

        usr_input = dialog_creator.StringInput().get_input_locale("select_gv")
        if usr_input is None:
            return None
        chunks = usr_input.split(" ")

        versions_selected: list[int] = []
        for chunk in chunks:
            parts = chunk.split("-")
            if len(parts) == 2:
                min = parts[0]
                max = parts[1]

                v1 = core.GameVersion.from_string(min)
                v2 = core.GameVersion.from_string(max)

                for v in range(v1.game_version, v2.game_version + 1):
                    versions_selected.append(v)
            else:
                v = core.GameVersion.from_string(chunk)
                versions_selected.append(v.game_version)

        valid_versions: set[int] = set()
        for version in versions_selected:
            if version in versions_set:
                valid_versions.add(version)

        if not valid_versions:
            color.ColoredText.localize("no_valid_gvs_entered")

        cats: list[core.Cat] = []
        for cat in self.save_file.cats.cats:
            row = unitbuy.get_unit_buy(cat.id)
            if row is None:
                continue
            if row.game_version in valid_versions:
                cats.append(cat)

        return cats

    def select_rarity(self) -> list[core.Cat] | None:
        rarity_names = self.save_file.cats.get_rarity_names(self.save_file)
        rarity_ids, _ = dialog_creator.ChoiceInput(
            rarity_names, rarity_names, [], {}, "select_rarity"
        ).multiple_choice()
        if rarity_ids is None:
            return None
        cats: list[core.Cat] = []
        for rarity_id in rarity_ids:
            rarity_cats = self.get_cats_rarity(rarity_id)
            cats = list(set(cats + rarity_cats))
        return cats

    def select_name(self) -> list[core.Cat] | None:
        usr_name = dialog_creator.StringInput().get_input_locale("enter_name", {})
        if usr_name is None:
            return []
        cats = self.get_cats_name(usr_name)
        if not cats:
            color.ColoredText.localize("no_cats_found_name", name=usr_name)
            return None
        cat_names: list[str] = []
        cat_list: list[core.Cat] = []
        for cat in cats:
            names = cat.get_names_cls(self.save_file)
            if not names:
                names = [str(cat.id)]
            for name in names:
                if usr_name.lower() in name.lower():
                    cat_names.append(name)
                    cat_list.append(cat)
                    break
        if len(cat_names) == 1:
            color.ColoredText(f"<@t>{cat_names[0]}</>")
        cat_option_ids, _ = dialog_creator.ChoiceInput(
            cat_names, cat_names, [], {}, "select_name"
        ).multiple_choice()
        if cat_option_ids is None:
            return None
        cats_selected: list[core.Cat] = []
        for cat_option_id in cat_option_ids:
            cats_selected.append(cat_list[cat_option_id])
        return cats_selected

    def select_obtainable(self) -> list[core.Cat] | None:
        return self.get_cats_obtainable()

    def select_gatya_banner_name(self) -> list[int] | None:

        filter_down = dialog_creator.YesNoInput().get_input_once("filter_down_q_gatya")
        if filter_down is None:
            return None

        all_names = core.GatyaInfos(self.save_file).get_all_names()
        ids = list(all_names.keys())
        ids.sort()
        names: list[str] = []
        for id in ids:
            names.append(all_names[id])
        new_names: list[str] = []
        new_ids: list[int] = []

        unknown_name = core.core_data.local_manager.get_key("unknown_banner")

        if filter_down:
            ids.reverse()
            for id in ids:
                name = all_names[id]
                if name in new_names or name == unknown_name:
                    continue
                new_names.append(name)
                new_ids.append(id)
            new_ids.reverse()
            new_names.reverse()
        else:
            new_names = names
            new_ids = ids

        ids = new_ids

        formatted_names: list[str] = []

        for name in new_names:
            formatted_name = core.core_data.local_manager.get_key(
                "banner_txt", name=name
            )
            formatted_names.append(formatted_name)
        gatya_option_ids, _ = dialog_creator.ChoiceInput.from_reduced(
            formatted_names,
            ints=ids,
            dialog="select_gatya_banner",
            start_index=0,
        ).multiple_choice(False)
        if gatya_option_ids is None:
            return None
        gatya_ids: list[int] = []
        for gatya_option_id in gatya_option_ids:
            gatya_ids.append(ids[gatya_option_id])

        return gatya_ids

    def select_gatya_banner(self) -> list[core.Cat] | None:
        gset = self.save_file.gatya.read_gatya_data_set(self.save_file).gatya_data_set
        if gset is None:
            return None

        by_id = dialog_creator.ChoiceInput.from_reduced(
            ["by_id", "by_name"], dialog="gatya_by_id_q"
        ).single_choice()
        if by_id is None:
            return None

        if by_id == 1:
            gatya_ids = dialog_creator.RangeInput(len(gset) - 1).get_input_locale(
                "select_gatya_banner", {}
            )
        else:
            gatya_ids = self.select_gatya_banner_name()
        if gatya_ids is None:
            return None
        cats: list[core.Cat] = []
        for gatya_id in gatya_ids:
            gatya_cats = self.get_cats_gatya_banner(gatya_id)
            if gatya_cats is None:
                continue
            cats = list(set(cats + gatya_cats))
        return cats

    def unlock_cats(self, cats: list[core.Cat]):
        cats = self.get_save_cats(cats)
        for cat in cats:
            cat.unlock(self.save_file)
        color.ColoredText.localize("unlock_success")

    def remove_cats(self, cats: list[core.Cat]):
        reset = core.core_data.config.get_bool(core.ConfigKey.RESET_CAT_DATA)
        cats = self.get_save_cats(cats)
        for cat in cats:
            cat.remove(reset=reset, save_file=self.save_file)
        color.ColoredText.localize("remove_success")

    def get_save_cats(self, cats: list[core.Cat]):
        ct_cats: list[core.Cat] = []
        for cat in cats:
            ct = self.save_file.cats.get_cat_by_id(cat.id)
            if ct is None:
                continue
            ct_cats.append(ct)
        return ct_cats

    def true_form_cats(self, cats: list[core.Cat], force: bool = False):
        cats = self.get_save_cats(cats)
        set_current_forms = core.core_data.config.get_bool(
            core.ConfigKey.SET_CAT_CURRENT_FORMS
        )
        self.save_file.cats.true_form_cats(
            self.save_file, cats, force, set_current_forms
        )
        color.ColoredText.localize("true_form_success")

    def fourth_form_cats(self, cats: list[core.Cat], force: bool = False):
        cats = self.get_save_cats(cats)
        set_current_forms = core.core_data.config.get_bool(
            core.ConfigKey.SET_CAT_CURRENT_FORMS
        )
        self.save_file.cats.fourth_form_cats(
            self.save_file, cats, force, set_current_forms
        )
        color.ColoredText.localize("fourth_form_success")

    def remove_true_form_cats(self, cats: list[core.Cat]):
        cats = self.get_save_cats(cats)
        for cat in cats:
            cat.remove_true_form()
        color.ColoredText.localize("remove_true_form_success")

    def remove_fourth_form_cats(self, cats: list[core.Cat]):
        cats = self.get_save_cats(cats)
        for cat in cats:
            cat.remove_fourth_form()
        color.ColoredText.localize("remove_fourth_form_success")

    def upgrade_cats(self, cats: list[core.Cat]):
        cats = self.get_save_cats(cats)
        if not cats:
            return
        if len(cats) == 1:
            option_id = 0
        else:
            options: list[str] = [
                "upgrade_individual",
                "upgrade_all",
            ]
            option_id = dialog_creator.ChoiceInput(
                options, options, [], {}, "upgrade_cats_select_mod", True
            ).single_choice()
            if option_id is None:
                return
            option_id -= 1
        success = False
        if option_id == 0:
            for cat in cats:
                names = cat.get_names_cls(self.save_file)
                if not names:
                    names = [str(cat.id)]
                color.ColoredText.localize(
                    "selected_cat_upgrades",
                    name=names[0],
                    id=cat.id,
                    base_level=cat.upgrade.base + 1,
                    plus_level=cat.upgrade.plus,
                )
                power_up = core.PowerUpHelper(cat, self.save_file)
                upgrade, should_exit = core.Upgrade.get_user_upgrade(
                    power_up.get_max_possible_base() - 1,
                    power_up.get_max_possible_plus(),
                )
                if should_exit:
                    return
                if upgrade is not None:
                    power_up.reset_upgrade()
                    power_up.upgrade_by(upgrade.base)
                    cat.set_upgrade(self.save_file, upgrade, True)
                    color.ColoredText.localize(
                        "selected_cat_upgraded",
                        name=names[0],
                        id=cat.id,
                        base_level=cat.upgrade.base + 1,
                        plus_level=cat.upgrade.plus,
                    )
                    success = True
        else:
            power_up = core.PowerUpHelper(cats[0], self.save_file)
            upgrade, should_exit = core.Upgrade.get_user_upgrade(
                power_up.get_max_max_base_upgrade_level() - 1,
                power_up.get_max_max_plus_upgrade_level(),
            )
            if upgrade is None or should_exit:
                return
            success = True
            for cat in cats:
                power_up = core.PowerUpHelper(cat, self.save_file)
                power_up.reset_upgrade()
                power_up.upgrade_by(upgrade.base)
                cat.set_upgrade(self.save_file, upgrade, True)
        if success:
            color.ColoredText.localize("upgrade_success")

    def remove_talents_cats(self, cats: list[core.Cat]):
        for cat in cats:
            if cat.talents is None:
                continue
            for talent in cat.talents:
                talent.level = 0
        color.ColoredText.localize("talents_remove_success")

    def unlock_cat_guide(self, cats: list[core.Cat]):
        for cat in cats:
            if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
                cat.unlock(self.save_file)
            cat.catguide_collected = True
        color.ColoredText.localize("unlock_cat_guide_success")

    def remove_cat_guide(self, cats: list[core.Cat]):
        for cat in cats:
            cat.catguide_collected = False
        color.ColoredText.localize("remove_cat_guide_success")

    def upgrade_talents_cats(self, cats: list[core.Cat]):
        cats = self.get_save_cats(cats)
        if not cats:
            return
        gdg = core.core_data.get_game_data_getter(self.save_file)
        is_good_version = gdg.does_save_version_match(self.save_file)
        if not is_good_version:
            data_version = gdg.version
            if data_version is None:
                color.ColoredText.localize("no_data_version")
                return
            color.ColoredText.localize(
                "talents_version_warning",
                save_version=self.save_file.game_version.to_string(),
                data_version=data_version,
            )
            should_stay = dialog_creator.YesNoInput().get_input_once("continue_q")
            if not should_stay:
                return

        if len(cats) == 1:
            option_id = 0
        else:
            options: list[str] = [
                "talents_individual",
                "talents_all",
            ]
            option_id = dialog_creator.ChoiceInput(
                options, options, [], {}, "upgrade_talents_select_mod", True
            ).single_choice()
            if option_id is None:
                return
            option_id -= 1

        talent_data = self.save_file.cats.read_talent_data(self.save_file)
        if talent_data is None:
            return
        if option_id == 0:
            for cat in cats:
                if cat.talents is None:
                    continue
                names = cat.get_names_cls(self.save_file)
                if not names:
                    names = [str(cat.id)]
                color.ColoredText.localize(
                    "selected_cat",
                    name=names[0],
                    id=cat.id,
                )
                data = talent_data.get_cat_talents(cat)
                if data is None:
                    color.ColoredText.localize("no_talent_data", id=cat.id)
                    continue
                if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
                    cat.unlock(self.save_file)
                talent_names, max_levels, current_levels, ids = data
                values = dialog_creator.MultiEditor.from_reduced(
                    "talents",
                    talent_names,
                    current_levels,
                    max_levels,
                    group_name_localized=True,
                ).edit()
                current_levels = values
                for i, id in enumerate(ids):
                    talent = cat.get_talent_from_id(id)
                    if talent is None:
                        continue
                    talent.level = current_levels[i]
        else:
            for cat in cats:
                if cat.talents is None:
                    continue
                data = talent_data.get_cat_talents(cat)
                if data is None:
                    continue
                if core.core_data.config.get_bool(core.ConfigKey.UNLOCK_CAT_ON_EDIT):
                    cat.unlock(self.save_file)
                talent_names, max_levels, current_levels, ids = data
                for i, id in enumerate(ids):
                    talent = cat.get_talent_from_id(id)
                    if talent is None:
                        continue
                    talent.level = max_levels[i]
        color.ColoredText.localize("talents_success")

    @staticmethod
    def edit_cats(save_file: core.SaveFile):
        cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        while True:
            should_exit, current_cats = cat_editor.run_edit_cats(current_cats)
            if should_exit:
                break

    @staticmethod
    def unlock_remove_cats_run(
        save_file: core.SaveFile,
        current_cats: list[core.Cat] | None = None,
        cat_editor: CatEditor | None = None,
    ):
        if cat_editor is None or current_cats is None:
            cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        choice = dialog_creator.ChoiceInput(
            ["unlock_cats", "remove_cats"],
            ["unlock_cats", "remove_cats"],
            [],
            {},
            "unlock_remove_q",
            True,
            remove_alias=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            cat_editor.unlock_cats(current_cats)
        elif choice == 1:
            cat_editor.remove_cats(current_cats)
        CatEditor.set_rank_up_sale(save_file)

    @staticmethod
    def true_form_remove_form_cats_run(
        save_file: core.SaveFile,
        current_cats: list[core.Cat] | None = None,
        cat_editor: CatEditor | None = None,
    ):
        if cat_editor is None or current_cats is None:
            cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        choice = dialog_creator.ChoiceInput.from_reduced(
            ["true_form_cats", "remove_true_form_cats"],
            dialog="true_form_remove_form_q",
            single_choice=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            cat_editor.true_form_cats(current_cats)
        elif choice == 1:
            cat_editor.remove_true_form_cats(current_cats)

    @staticmethod
    def fourth_form_remove_form_cats_run(
        save_file: core.SaveFile,
        current_cats: list[core.Cat] | None = None,
        cat_editor: CatEditor | None = None,
    ):
        if cat_editor is None or current_cats is None:
            cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        choice = dialog_creator.ChoiceInput.from_reduced(
            ["fourth_form_cats", "remove_fourth_form_cats"],
            dialog="fourth_form_remove_form_q",
            single_choice=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            cat_editor.fourth_form_cats(current_cats)
        elif choice == 1:
            cat_editor.remove_fourth_form_cats(current_cats)

    @staticmethod
    def force_true_form_cats_run(save_file: core.SaveFile):
        color.ColoredText.localize("force_true_form_cats_warning")
        cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        cat_editor.true_form_cats(current_cats, force=True)

    @staticmethod
    def force_fourth_form_cats_run(save_file: core.SaveFile):
        color.ColoredText.localize("force_fourth_form_cats_warning")
        cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        cat_editor.fourth_form_cats(current_cats, force=True)

    @staticmethod
    def upgrade_cats_run(save_file: core.SaveFile):
        cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        cat_editor.upgrade_cats(current_cats)
        CatEditor.set_rank_up_sale(save_file)

    @staticmethod
    def upgrade_talents_remove_talents_cats_run(
        save_file: core.SaveFile,
        current_cats: list[core.Cat] | None = None,
        cat_editor: CatEditor | None = None,
    ):
        if cat_editor is None or current_cats is None:
            cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        choice = dialog_creator.ChoiceInput(
            ["upgrade_talents_cats", "remove_talents_cats"],
            ["upgrade_talents_cats", "remove_talents_cats"],
            [],
            {},
            "upgrade_talents_remove_talents_q",
            True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            cat_editor.upgrade_talents_cats(current_cats)
        elif choice == 1:
            cat_editor.remove_talents_cats(current_cats)

    @staticmethod
    def unlock_cat_guide_remove_guide_run(
        save_file: core.SaveFile,
        current_cats: list[core.Cat] | None = None,
        cat_editor: CatEditor | None = None,
    ):
        if cat_editor is None or current_cats is None:
            cat_editor, current_cats = CatEditor.from_save_file(save_file)
        if cat_editor is None:
            return
        choice = dialog_creator.ChoiceInput(
            ["unlock_cat_guide", "remove_cat_guide"],
            ["unlock_cat_guide", "remove_cat_guide"],
            [],
            {},
            "unlock_cat_guide_remove_guide_q",
            True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            cat_editor.unlock_cat_guide(current_cats)
        elif choice == 1:
            cat_editor.remove_cat_guide(current_cats)

    @staticmethod
    def from_save_file(
        save_file: core.SaveFile,
    ) -> tuple[CatEditor | None, list[core.Cat]]:
        cat_editor = CatEditor(save_file)
        stop = False
        cats = []
        while not stop:
            current_cats, finished = cat_editor.select(cats)
            cats = current_cats
            cat_editor.print_selected_cats(cats)
            if finished:
                stop = True
                continue
            finished = dialog_creator.YesNoInput().get_input_once(
                "finished_cats_selection"
            )
            if finished is None:
                return None, []
            stop = finished
        return cat_editor, cats

    def run_edit_cats(
        self,
        cats: list[core.Cat],
    ) -> tuple[bool, list[core.Cat]]:
        self.print_selected_cats(cats)
        options: list[str] = [
            "select_cats_again",
            "unlock_remove_cats",
            "upgrade_cats",
            "true_form_remove_form_cats",
            "force_true_form_cats",
            "fourth_form_remove_form_cats",
            "force_fourth_form_cats",
            "upgrade_talents_remove_talents_cats",
            "unlock_remove_cat_guide",
            "finish_edit_cats",
        ]
        option_id = dialog_creator.ChoiceInput(
            options,
            options,
            [],
            {},
            "select_edit_cats_option",
            True,
            remove_alias=True,
        ).single_choice()
        if option_id is None:
            return False, cats
        option_id -= 1
        if option_id == 0:
            cats_, _ = self.select(cats, False)
            cats = cats_
        elif option_id == 1:
            self.unlock_remove_cats_run(self.save_file, cats, self)
        elif option_id == 2:
            self.upgrade_cats(cats)
        elif option_id == 3:
            self.true_form_remove_form_cats_run(self.save_file, cats, self)
        elif option_id == 4:
            color.ColoredText.localize("force_true_form_cats_warning")
            self.true_form_cats(cats, force=True)
        elif option_id == 5:
            self.fourth_form_remove_form_cats_run(self.save_file, cats, self)
        elif option_id == 6:
            color.ColoredText.localize("force_fourth_form_cats_warning")
            self.fourth_form_cats(cats, force=True)
        elif option_id == 7:
            self.upgrade_talents_remove_talents_cats_run(self.save_file, cats, self)
        elif option_id == 8:
            self.unlock_cat_guide_remove_guide_run(self.save_file, cats, self)
        CatEditor.set_rank_up_sale(self.save_file)
        if option_id == 9:
            return True, cats
        return False, cats

    @staticmethod
    def set_rank_up_sale(save_file: core.SaveFile):
        save_file.rank_up_sale_value = 0x7FFFFFFF



from __future__ import annotations
from bcsfe.cli import dialog_creator, main, color, file_dialog
from bcsfe import core


class ServerCLI:
    def __init__(self):
        pass

    def download_save(
        self,
    ) -> tuple[core.Path, core.CountryCode] | None:
        transfer_code = dialog_creator.StringInput().get_input_locale_while(
            "enter_transfer_code", {}
        )
        if transfer_code is None:
            return None
        confirmation_code = dialog_creator.StringInput().get_input_locale_while(
            "enter_confirmation_code", {}
        )
        if confirmation_code is None:
            return None
        cc = core.CountryCode.select()
        if cc is None:
            return None
        gv = core.GameVersion(120200)  # not important

        color.ColoredText.localize(
            "downloading_save_file",
            transfer_code=transfer_code,
            confirmation_code=confirmation_code,
            country_code=cc,
        )

        server_handler, result = core.ServerHandler.from_codes(
            transfer_code,
            confirmation_code,
            cc,
            gv,
        )
        if server_handler is None and result is not None:
            color.ColoredText.localize("invalid_codes_error")
            if dialog_creator.YesNoInput().get_input_once(
                "display_response_debug_info_q"
            ):
                if result.response is not None:
                    color.ColoredText.localize(
                        "response_text_display",
                        url=result.url,
                        request_headers=result.headers,
                        request_body=result.data,
                        response_headers=result.response.headers,
                        response_body=result.response.text,
                    )
            return
        if server_handler is None:
            return

        save_file = server_handler.save_file
        if file_dialog.FileDialog().filedialog is None:
            path = core.SaveFile.get_saves_path().add("SAVE_DATA")
        else:
            path = main.Main().save_save_dialog(save_file)
        if path is None:
            return None

        save_file.to_file(path)

        color.ColoredText.localize("save_downloaded", path=path.to_str())

        return path, cc



from __future__ import annotations

from bcsfe import core


class MapOptionLine:
    def __init__(
        self,
        map_id: int,
        crown_count: int,
        crown_mults: list[int],
        guerrilla_set: int,
        reset_type: int,
        one_time_display: bool,
        display_order: int,
        interval: int,
        challenge_flag: bool,
        difficulty_mask: int,
        hide_after_clear: bool,
        name: str,
    ):
        self.map_id = map_id
        self.crown_count = crown_count
        self.crown_mults = crown_mults
        self.guerrilla_set = guerrilla_set
        self.reset_type = reset_type
        self.one_time_display = one_time_display
        self.display_order = display_order
        self.interval = interval
        self.challenge_flag = challenge_flag
        self.difficulty_mask = difficulty_mask
        self.hide_after_clear = hide_after_clear
        self.name = name

    @staticmethod
    def from_line(line: core.Row) -> MapOptionLine:
        return MapOptionLine(
            line.next_int(),
            line.next_int(),
            [line.next_int() for _ in range(4)],
            line.next_int(),
            line.next_int(),
            line.next_bool(),
            line.next_int(),
            line.next_int(),
            line.next_bool(),
            line.next_int(),
            line.next_bool(),
            line.next_str(),
        )


class MapOption:
    def __init__(self, maps: dict[int, MapOptionLine]):
        self.maps = maps

    @staticmethod
    def from_csv(csv: core.CSV) -> MapOption:
        data: dict[int, MapOptionLine] = {}

        for line in csv.lines[1:]:  # skip headers
            item = MapOptionLine.from_line(line)
            data[item.map_id] = item

        return MapOption(data)

    @staticmethod
    def from_save(save_file: core.SaveFile) -> MapOption | None:
        gdg = core.core_data.get_game_data_getter(save_file)
        data = gdg.download("DataLocal", "Map_option.csv")
        if data is None:
            return None

        csv = core.CSV(data)

        return MapOption.from_csv(csv)

    def get_map(self, map_id: int) -> MapOptionLine | None:
        return self.maps.get(map_id)



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class CatShrine:
    def __init__(
        self,
        unknown: bool,
        stamp_1: float,
        stamp_2: float,
        shrine_gone: bool,
        flags: list[int],
        xp_offering: int,
    ):
        self.unknown = unknown
        self.stamp_1 = stamp_1
        self.stamp_2 = stamp_2
        self.shrine_gone = shrine_gone
        self.flags = flags
        self.xp_offering = xp_offering
        self.dialogs = 0

    @staticmethod
    def init() -> CatShrine:
        return CatShrine(False, 0.0, 0.0, False, [], 0)

    @staticmethod
    def read(stream: core.Data) -> CatShrine:
        unknown = stream.read_bool()
        stamp_1 = stream.read_double()
        stamp_2 = stream.read_double()
        shrine_gone = stream.read_bool()
        flags = stream.read_byte_list(length=stream.read_byte())
        xp_offering = stream.read_long()
        return CatShrine(unknown, stamp_1, stamp_2, shrine_gone, flags, xp_offering)

    def write(self, stream: core.Data):
        stream.write_bool(self.unknown)
        stream.write_double(self.stamp_1)
        stream.write_double(self.stamp_2)
        stream.write_bool(self.shrine_gone)
        stream.write_byte(len(self.flags))
        stream.write_byte_list(self.flags, write_length=False)
        stream.write_long(self.xp_offering)

    def read_dialogs(self, stream: core.Data):
        self.dialogs = stream.read_int()

    def write_dialogs(self, stream: core.Data):
        stream.write_int(self.dialogs)

    def serialize(self) -> dict[str, Any]:
        return {
            "unknown": self.unknown,
            "stamp_1": self.stamp_1,
            "stamp_2": self.stamp_2,
            "shrine_gone": self.shrine_gone,
            "flags": self.flags,
            "xp_offering": self.xp_offering,
            "dialogs": self.dialogs,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> CatShrine:
        shrine = CatShrine(
            data.get("unknown", False),
            data.get("stamp_1", 0.0),
            data.get("stamp_2", 0.0),
            data.get("shrine_gone", False),
            data.get("flags", []),
            data.get("xp_offering", 0),
        )
        shrine.dialogs = data.get("dialogs", 0)
        return shrine

    def __repr__(self):
        return (
            f"CatShrine("
            f"unknown={self.unknown}, "
            f"stamp_1={self.stamp_1}, "
            f"stamp_2={self.stamp_2}, "
            f"shrine_gone={self.shrine_gone}, "
            f"flags={self.flags}, "
            f"xp_offering={self.xp_offering}, "
            f"dialogs={self.dialogs}"
            f")"
        )

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def edit_catshrine(save_file: core.SaveFile):
        shrine = save_file.cat_shrine
        options = [
            "shrine_level",
            "shrine_xp",
            "make_catshrine_appear",
            "make_catshrine_disappear",
        ]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="cat_shrine_choice_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return
        choice -= 1

        if choice == 2:
            shrine.shrine_gone = False

            shrine.stamp_1 = 0.0
            shrine.stamp_2 = 0.0
            color.ColoredText.localize("cat_shrine_edited")
            return
        elif choice == 3:
            shrine.shrine_gone = True
            color.ColoredText.localize("cat_shrine_edited")
            return

        data = core.core_data.get_cat_shrine_levels(save_file)

        xp = shrine.xp_offering
        level = data.get_level_from_xp(xp)

        color.ColoredText.localize("current_shrine_xp_level", level=level, xp=xp)

        if choice == 0:
            max_level = data.get_max_level()
            if max_level is None:
                return
            level = dialog_creator.IntInput(
                min=1, max=max_level
            ).get_input_locale_while("shrine_level_dialog", {"max_level": max_level})
            if level is None:
                return
            shrine.xp_offering = data.get_xp_from_level(level)
        elif choice == 1:
            max_xp = data.get_max_xp()
            if max_xp is None:
                return
            xp = dialog_creator.IntInput(min=0, max=max_xp).get_input_locale_while(
                "shrine_xp_dialog", {"max_xp": max_xp}
            )
            if xp is None:
                return
            shrine.xp_offering = xp

        xp = shrine.xp_offering
        if xp is None:
            return
        level = data.get_level_from_xp(xp)
        if level is None:
            return

        shrine.dialogs = level - 1
        shrine.shrine_gone = False
        shrine.stamp_1 = 0.0
        shrine.stamp_2 = 0.0

        color.ColoredText.localize("current_shrine_xp_level", level=level, xp=xp)

        color.ColoredText.localize("cat_shrine_edited")


class CatShrineLevels:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.boundaries = self.get_boundaries()

    def get_boundaries(self) -> list[int] | None:
        file_name = "jinja_level.csv"
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", file_name)
        if data is None:
            return None
        csv = core.CSV(
            data,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        boundaries: list[int] = []
        counter = 0
        for row in csv:
            xp = row[0].to_int()
            counter += xp
            boundaries.append(counter)

        return boundaries

    def get_level_from_xp(self, xp: int) -> int | None:
        if self.boundaries is None:
            return None
        for i, boundary in enumerate(self.boundaries):
            if xp < boundary:
                return i + 1
        return len(self.boundaries)

    def get_xp_from_level(self, level: int) -> int | None:
        if self.boundaries is None:
            return None
        if level < 1:
            return 0
        if level > len(self.boundaries):
            return self.get_max_xp()
        return self.boundaries[level - 2]

    def get_max_level(self) -> int | None:
        if self.boundaries is None:
            return None
        return len(self.boundaries)

    def get_max_xp(self) -> int | None:
        if self.boundaries is None:
            return None
        return max(self.boundaries)



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import edits


class Stage:
    def __init__(self, clear_times: int):
        self.clear_times = clear_times

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(data: core.Data) -> Stage:
        clear_times = data.read_short()
        return Stage(clear_times)

    def write(self, data: core.Data):
        data.write_short(self.clear_times)

    def serialize(self) -> int:
        return self.clear_times

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(
            data,
        )

    def __repr__(self):
        return f"Stage({self.clear_times})"

    def __str__(self):
        return self.__repr__()

    def clear_stage(self, clear_amount: int = 1, ensure_cleared_only: bool = False):
        if ensure_cleared_only:
            self.clear_times = self.clear_times or clear_amount
        else:
            self.clear_times = clear_amount

    def unclear_stage(self):
        self.clear_times = 0


class Chapter:
    def __init__(self, selected_stage: int, total_stages: int = 0):
        self.selected_stage = selected_stage
        self.clear_progress = 0
        self.stages: list[Stage] = [Stage.init() for _ in range(total_stages)]
        self.chapter_unlock_state = 0
        self.total_stages = 0

    def clear_stage(
        self,
        index: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        if overwrite_clear_progress:
            self.clear_progress = index + 1
        else:
            self.clear_progress = max(self.clear_progress, index + 1)
        self.stages[index].clear_stage(clear_amount, ensure_cleared_only)
        self.chapter_unlock_state = 3
        if index == self.total_stages - 1:
            return True
        return False

    def unclear_stage(self, index: int):
        self.clear_progress = min(self.clear_progress, index)
        self.stages[index].unclear_stage()
        return True

    @staticmethod
    def init(total_stages: int) -> Chapter:
        return Chapter(0, total_stages)

    @staticmethod
    def read_selected_stage(data: core.Data) -> Chapter:
        selected_stage = data.read_byte()
        return Chapter(selected_stage)

    def write_selected_stage(self, data: core.Data):
        data.write_byte(self.selected_stage)

    def read_clear_progress(self, data: core.Data):
        self.clear_progress = data.read_byte()

    def write_clear_progress(self, data: core.Data):
        data.write_byte(self.clear_progress)

    def read_stages(self, data: core.Data, total_stages: int):
        self.stages = [Stage.read(data) for _ in range(total_stages)]

    def write_stages(self, data: core.Data):
        for stage in self.stages:
            stage.write(data)

    def read_chapter_unlock_state(self, data: core.Data):
        self.chapter_unlock_state = data.read_byte()

    def write_chapter_unlock_state(self, data: core.Data):
        data.write_byte(self.chapter_unlock_state)

    def serialize(self) -> dict[str, Any]:
        return {
            "selected_stage": self.selected_stage,
            "clear_progress": self.clear_progress,
            "stages": [stage.serialize() for stage in self.stages],
            "chapter_unlock_state": self.chapter_unlock_state,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Chapter:
        chapter = Chapter(data.get("selected_stage", 0))
        chapter.clear_progress = data.get("clear_progress", 0)
        chapter.stages = [Stage.deserialize(stage) for stage in data.get("stages", [])]
        chapter.chapter_unlock_state = data.get("chapter_unlock_state", 0)
        return chapter

    def __repr__(self):
        return f"Chapter({self.selected_stage}, {self.clear_progress}, {self.stages}, {self.chapter_unlock_state})"

    def __str__(self):
        return self.__repr__()


class ChaptersStars:
    def __init__(self, chapters: list[Chapter]):
        self.chapters = chapters

    def clear_stage(
        self,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        finished = self.chapters[star].clear_stage(
            stage, clear_amount, overwrite_clear_progress, ensure_cleared_only
        )
        if finished:
            if star + 1 < len(self.chapters):
                self.chapters[star + 1].chapter_unlock_state = 1
        return finished

    def unclear_stage(self, star: int, stage: int):
        finished = self.chapters[star].unclear_stage(stage)
        if finished and star + 1 < len(self.chapters):
            for chapter in self.chapters[star + 1 :]:
                chapter.chapter_unlock_state = 0
        return finished

    @staticmethod
    def init(total_stages: int, total_stars: int) -> ChaptersStars:
        chapters = [Chapter.init(total_stages) for _ in range(total_stars)]
        return ChaptersStars(chapters)

    @staticmethod
    def read_selected_stage(data: core.Data, total_stars: int) -> ChaptersStars:
        chapters = [Chapter.read_selected_stage(data) for _ in range(total_stars)]
        return ChaptersStars(chapters)

    def write_selected_stage(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_selected_stage(data)

    def read_clear_progress(self, data: core.Data):
        for chapter in self.chapters:
            chapter.read_clear_progress(data)

    def write_clear_progress(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_clear_progress(data)

    def read_stages(self, data: core.Data, total_stages: int):
        for _ in range(total_stages):
            for chapter in self.chapters:
                chapter.stages.append(Stage.read(data))

    def write_stages(self, data: core.Data):
        for i in range(len(self.chapters[0].stages)):
            for chapter in self.chapters:
                chapter.stages[i].write(data)

    def read_chapter_unlock_state(self, data: core.Data):
        for chapter in self.chapters:
            chapter.read_chapter_unlock_state(data)

    def write_chapter_unlock_state(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_chapter_unlock_state(data)

    def serialize(self) -> list[dict[str, Any]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> ChaptersStars:
        chapters = [Chapter.deserialize(chapter) for chapter in data]
        return ChaptersStars(chapters)

    def __repr__(self):
        return f"ChaptersStars({self.chapters})"

    def __str__(self):
        return self.__repr__()


class GauntletChapters:
    def __init__(self, chapters: list[ChaptersStars], unknown: list[int]):
        self.chapters = chapters
        self.unknown = unknown

    def clear_stage(
        self,
        map: int,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        finished = self.chapters[map].clear_stage(
            star, stage, clear_amount, overwrite_clear_progress, ensure_cleared_only
        )
        if finished and map + 1 < len(self.chapters):
            self.chapters[map + 1].chapters[0].chapter_unlock_state = 1

        return finished

    def unclear_stage(self, map: int, star: int, stage: int) -> bool:
        finished = self.chapters[map].unclear_stage(star, stage)
        if finished and map + 1 < len(self.chapters) and star == 0:
            for chapter in self.chapters[map + 1].chapters:
                chapter.chapter_unlock_state = 0

        return finished

    @staticmethod
    def init() -> GauntletChapters:
        return GauntletChapters([], [])

    @staticmethod
    def read(data: core.Data) -> GauntletChapters:
        total_chapters = data.read_short()
        total_stages = data.read_byte()
        total_stars = data.read_byte()

        chapters = [
            ChaptersStars.read_selected_stage(data, total_stars)
            for _ in range(total_chapters)
        ]

        for chapter in chapters:
            chapter.read_clear_progress(data)

        for chapter in chapters:
            chapter.read_stages(data, total_stages)

        for chapter in chapters:
            chapter.read_chapter_unlock_state(data)

        unknown = [data.read_byte() for _ in range(total_chapters)]

        return GauntletChapters(chapters, unknown)

    def write(self, data: core.Data):
        data.write_short(len(self.chapters))
        try:
            data.write_byte(len(self.chapters[0].chapters[0].stages))
        except IndexError:
            data.write_byte(0)
        try:
            data.write_byte(len(self.chapters[0].chapters))
        except IndexError:
            data.write_byte(0)
        for chapter in self.chapters:
            chapter.write_selected_stage(data)

        for chapter in self.chapters:
            chapter.write_clear_progress(data)

        for chapter in self.chapters:
            chapter.write_stages(data)

        for chapter in self.chapters:
            chapter.write_chapter_unlock_state(data)

        for unknown in self.unknown:
            data.write_byte(unknown)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": [chapter.serialize() for chapter in self.chapters],
            "unknown": self.unknown,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> GauntletChapters:
        chapters = [
            ChaptersStars.deserialize(chapter) for chapter in data.get("chapters", [])
        ]
        return GauntletChapters(chapters, data.get("unknown", []))

    def __repr__(self):
        return f"Chapters({self.chapters}, {self.unknown})"

    def __str__(self):
        return self.__repr__()

    def get_total_stars(self, map: int) -> int:
        try:
            return len(self.chapters[map].chapters)
        except IndexError:
            return 0

    def get_total_stages(self, map: int, star: int) -> int:
        try:
            return len(self.chapters[map].chapters[star].stages)
        except IndexError:
            return 0

    @staticmethod
    def edit_gauntlets(save_file: core.SaveFile):
        gauntlets = save_file.gauntlets
        gauntlets.edit_chapters(save_file, "A", 24000)

    @staticmethod
    def edit_collab_gauntlets(save_file: core.SaveFile):
        gauntlets = save_file.collab_gauntlets
        gauntlets.edit_chapters(save_file, "CA", 27000)

    @staticmethod
    def edit_behemoth_culling(save_file: core.SaveFile):
        gauntlets = save_file.behemoth_culling
        gauntlets.edit_chapters(save_file, "Q", 31000)

    @staticmethod
    def edit_enigma_stages(save_file: core.SaveFile):
        save_file.enigma_clears.edit_chapters(save_file, "H", 25000)

    def edit_chapters(
        self, save_file: core.SaveFile, letter_code: str, base_index: int
    ):
        edits.map.edit_chapters(save_file, self, letter_code, base_index=base_index)

    def unclear_rest(self, stages: list[int], stars: int, id: int):
        if not stages:
            return
        for star in range(stars, self.get_total_stars(id)):
            for stage in range(max(stages), self.get_total_stages(id, star)):
                self.chapters[id].chapters[star].stages[stage].clear_times = 0
                self.chapters[id].chapters[star].clear_progress = 0

    def set_total_stages(self, map: int, total_stages: int):
        for chapter in self.chapters[map].chapters:
            chapter.total_stages = total_stages



from __future__ import annotations
from typing import Any
from bcsfe import core


class PurchasedPack:
    def __init__(self, purchased: bool):
        self.purchased = purchased

    @staticmethod
    def init() -> PurchasedPack:
        return PurchasedPack(False)

    @staticmethod
    def read(stream: core.Data) -> PurchasedPack:
        purchased = stream.read_bool()
        return PurchasedPack(purchased)

    def write(self, stream: core.Data):
        stream.write_bool(self.purchased)

    def serialize(self) -> bool:
        return self.purchased

    @staticmethod
    def deserialize(data: bool) -> PurchasedPack:
        return PurchasedPack(data)

    def __repr__(self) -> str:
        return f"PurchasedPack(purchased={self.purchased!r})"

    def __str__(self) -> str:
        return self.__repr__()


class PurchaseSet:
    def __init__(self, purchases: dict[str, PurchasedPack]):
        self.purchases = purchases

    @staticmethod
    def init() -> PurchaseSet:
        return PurchaseSet({})

    @staticmethod
    def read(stream: core.Data) -> PurchaseSet:
        total = stream.read_int()
        purchases: dict[str, PurchasedPack] = {}
        for _ in range(total):
            key = stream.read_string()
            purchases[key] = PurchasedPack.read(stream)
        return PurchaseSet(purchases)

    def write(self, stream: core.Data):
        stream.write_int(len(self.purchases))
        for key, purchase in self.purchases.items():
            stream.write_string(key)
            purchase.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            key: purchase.serialize()
            for key, purchase in self.purchases.items()
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> PurchaseSet:
        return PurchaseSet(
            {
                key: PurchasedPack.deserialize(purchase)
                for key, purchase in data.items()
            },
        )

    def __repr__(self) -> str:
        return f"PurchaseSet(purchases={self.purchases!r})"

    def __str__(self) -> str:
        return self.__repr__()


class Purchases:
    def __init__(self, purchases: dict[int, PurchaseSet]):
        self.purchases = purchases

    @staticmethod
    def init() -> Purchases:
        return Purchases({})

    @staticmethod
    def read(stream: core.Data) -> Purchases:
        total = stream.read_int()
        purchases: dict[int, PurchaseSet] = {}
        for _ in range(total):
            key = stream.read_int()
            purchases[key] = PurchaseSet.read(stream)

        return Purchases(purchases)

    def write(self, stream: core.Data):
        stream.write_int(len(self.purchases))
        for key, purchase in self.purchases.items():
            stream.write_int(key)
            purchase.write(stream)

    def serialize(self) -> dict[int, Any]:
        return {
            key: purchase.serialize()
            for key, purchase in self.purchases.items()
        }

    @staticmethod
    def deserialize(data: dict[int, Any]) -> Purchases:
        return Purchases(
            {
                key: PurchaseSet.deserialize(purchase)
                for key, purchase in data.items()
            },
        )

    def __repr__(self) -> str:
        return f"Purchases(purchases={self.purchases!r})"

    def __str__(self) -> str:
        return self.__repr__()


class ItemPack:
    def __init__(self, purchases: Purchases):
        self.purchases = purchases
        self.displayed_packs: dict[int, bool] = {}
        self.three_days_started: bool = False
        self.three_days_end_timestamp: float = 0.0

    @staticmethod
    def init() -> ItemPack:
        return ItemPack(Purchases.init())

    @staticmethod
    def read(stream: core.Data) -> ItemPack:
        return ItemPack(Purchases.read(stream))

    def write(self, stream: core.Data):
        self.purchases.write(stream)

    def read_displayed_packs(self, stream: core.Data) -> None:
        total = stream.read_int()
        displayed_packs: dict[int, bool] = {}
        for _ in range(total):
            key = stream.read_int()
            displayed_packs[key] = stream.read_bool()

        self.displayed_packs = displayed_packs

    def write_displayed_packs(self, stream: core.Data) -> None:
        stream.write_int(len(self.displayed_packs))
        for key, displayed in self.displayed_packs.items():
            stream.write_int(key)
            stream.write_bool(displayed)

    def read_three_days(self, stream: core.Data) -> None:
        self.three_days_started = stream.read_bool()
        self.three_days_end_timestamp = stream.read_double()

    def write_three_days(self, stream: core.Data) -> None:
        stream.write_bool(self.three_days_started)
        stream.write_double(self.three_days_end_timestamp)

    def serialize(self) -> dict[str, Any]:
        return {
            "purchases": self.purchases.serialize(),
            "displayed_packs": self.displayed_packs,
            "three_days_started": self.three_days_started,
            "three_days_end_timestamp": self.three_days_end_timestamp,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ItemPack:
        item_pack = ItemPack(Purchases.deserialize(data.get("purchases", {})))
        item_pack.displayed_packs = data.get("displayed_packs", {})
        item_pack.three_days_started = data.get("three_days_started", False)
        item_pack.three_days_end_timestamp = data.get(
            "three_days_end_timestamp", 0.0
        )
        return item_pack

    def __repr__(self) -> str:
        return f"ItemPack(purchases={self.purchases!r}, displayed_packs={self.displayed_packs!r}, three_days_started={self.three_days_started!r}, three_days_end_timestamp={self.three_days_end_timestamp!r})"

    def __str__(self) -> str:
        return self.__repr__()
